#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on January 10, 2017
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from traits.api import (
    Bool, Dict, HasStrictTraits, Instance, Property, on_trait_change)

from karabo.native import Hash
from karabogui.binding.api import (
    NO_CLASS_STATUSES, BindingRoot, DeviceClassProxy, DeviceProxy,
    ProjectDeviceProxy, ProxyStatus, apply_configuration,
    apply_default_configuration, build_binding)

from .device_tree import DeviceSystemTree
from .project_device import ProjectDeviceInstance
from .system_tree import SystemTree


class SystemTopology(HasStrictTraits):
    """The Karabo System Topology

    This is a data structure containing all of the online devices and device
    servers in a running Karabo system. It is kept continually up-to-date by
    the connection to the GUI server.
    """

    # A high-level version of `_system_hash`. Used by GUI views.
    system_tree = Instance(SystemTree)

    # A high-level version for device folded structure
    device_tree = Instance(DeviceSystemTree)

    # True if the system is online, False otherwise
    online = Property(Bool)

    # Mapping of (server id, class id) -> DeviceClassProxy
    _class_proxies = Dict

    # Mapping of server_id-> {class_id: Schema}
    _class_schemas = Dict

    # Tracking of requested Class Schemas for {server id: set()} -> class id
    _requested_classes = Dict

    # Mapping of device_id -> DeviceProxy
    _device_proxies = Dict

    # Mapping of device_id -> ProjectDeviceInstance
    _project_devices = Dict

    # Mapping of (server id, class id) -> {device_id: ProjectDeviceProxy}
    _project_device_proxies = Dict

    # A Hash instance holding the entire current topology
    _system_hash = Instance(Hash, args=())

    # Required by the singleton system (see ``karabogui.singleton.api``)
    def __init__(self, parent=None):
        super().__init__()

    def clear(self):
        """Clear all saved devices and classes"""
        self.clear_project_devices()
        self.system_tree.clear_all()
        self.device_tree.clear_all()

        self._system_hash.clear()
        self._class_proxies.clear()
        self._class_schemas.clear()
        self._device_proxies.clear()
        self._requested_classes.clear()

    def clear_project_devices(self):
        """Called by project model on closing project"""
        self._project_devices.clear()
        self._project_device_proxies.clear()

    def get_attributes(self, topology_path):
        """Return the attributes of a given node in the `_system_hash`"""
        try:
            return self._system_hash[topology_path, ...]
        except KeyError:
            return None

    def get_class(self, server_id, class_id):
        """Return the class proxy (DeviceClassProxy) for a given
        class on a given server

        This function will request the class schema if not available
        """
        key = (server_id, class_id)
        proxy = self._class_proxies.get(key)
        if proxy is None:
            binding = BindingRoot(class_id=class_id)
            proxy = DeviceClassProxy(server_id=server_id, binding=binding)
            self._class_proxies[key] = proxy
            path = 'server' + '.' + server_id
            attrs = self.get_attributes(path)
            if attrs and class_id in attrs.get('deviceClasses', ()):
                # NOTE: The server is online and has the device class we want
                # but we only request if it was not previously requested!
                request = class_id not in self._requested_classes.setdefault(
                    server_id, set())
                if request:
                    self._requested_classes[server_id].add(class_id)
                    proxy.refresh_schema()
                else:
                    proxy.status = ProxyStatus.REQUESTED

        return proxy

    def get_device(self, device_id, request=True):
        """Return the online version proxy (DeviceProxy) for a given device

        :param request: Request a schema for the freshly created online proxy
                        This is not required for lazy building, however, the
                        default is `True`.
        """
        proxy = self._device_proxies.get(device_id)
        if proxy is None:
            binding = BindingRoot()
            proxy = DeviceProxy(device_id=device_id, binding=binding)
            self._device_proxies[device_id] = proxy

            attrs = self._get_device_attributes(device_id)
            if attrs is not None:
                # We are dealing with an online device. Hence we assign
                # the class id and server id from the system hash and
                # if desired, request a schema.
                binding.class_id = attrs['classId']
                proxy.server_id = attrs['serverId']
                if request:
                    proxy.refresh_schema()
                elif proxy.status is ProxyStatus.OFFLINE:
                    # We did not request a schema, but we have been freshly
                    # created and are ONLINE
                    proxy.status = ProxyStatus.ONLINE
            # Get the system topology node, if it's there
            node = self.system_tree.get_instance_node(device_id)
            if node:
                proxy.topology_node = node

        return proxy

    def get_project_device(self, device_id, server_id='', class_id='',
                           init_config=None):
        """Return a ``ProjectDeviceInstance`` for a device on a specific
        server.

        The ProjectDeviceInstance holds two proxies, when the device is online
        it uses DeviceProxy, when the device is offline, it uses
        ProjectDeviceProxy (which is a modified version of ClassDeviceProxy)
        """
        if device_id not in self._project_devices:
            if class_id == '' or server_id == '':
                msg = ('Project device with id "{}" was requested without '
                       'specifying the class_id or server_id! Perhaps a '
                       'device exists in a scene workflow but not in an open '
                       'project?')
                raise RuntimeError(msg.format(device_id))

            instance = ProjectDeviceInstance(device_id, server_id, class_id)
            instance.set_project_config_hash(init_config)
            self._project_devices[device_id] = instance
        else:
            instance = self._project_devices[device_id]
            instance.rename(
                device_id=device_id, class_id=class_id, server_id=server_id)

        return self._project_devices[device_id]

    def delete_project_device(self, device_id):
        return self._project_devices.pop(device_id, None)

    def get_project_device_proxy(self, device_id, server_id, class_id):
        """Return the project device proxy (offline version of a device in a
        certain project) for a given class on a given server

        Offline devices are always created lazily with an empty `BindingRoot`.
        """
        key = (server_id, class_id)
        mapping = self._project_device_proxies.setdefault(key, {})
        proxy = mapping.get(device_id)
        if proxy is None:
            binding = BindingRoot(class_id=class_id)
            proxy = ProjectDeviceProxy(device_id=device_id,
                                       server_id=server_id,
                                       binding=binding)
            mapping[device_id] = proxy

        return proxy

    def get_schema(self, server_id, class_id):
        """Return the schema for a given device class on a server.
        """
        classes = self._class_schemas.setdefault(server_id, {})
        return classes.get(class_id)

    def ensure_proxy_class_schema(self, device_id, server_id, class_id):
        """Ensure the class schema of a project device proxy if necessary"""
        key = (server_id, class_id)
        mapping = self._project_device_proxies.setdefault(key, {})
        proxy = mapping.get(device_id)
        if proxy is not None and not len(proxy.binding.value) > 0:
            schema = self.get_schema(server_id, class_id)
            if schema is not None:
                build_binding(schema, existing=proxy.binding)
            elif proxy.status not in NO_CLASS_STATUSES:
                # We only request if it was not previously requested!
                request = class_id not in self._requested_classes.setdefault(
                    server_id, set())
                if request:
                    self._requested_classes[server_id].add(class_id)
                    proxy.refresh_schema()
                else:
                    proxy.status = ProxyStatus.REQUESTED

    def remove_project_device_proxy(self, device_id, server_id, class_id):
        """Remove the project device proxy for a given instance.
        """
        key = (server_id, class_id)
        mapping = self._project_device_proxies.get(key, {})
        mapping.pop(device_id, None)
        if not mapping:
            self._project_device_proxies.pop(key, None)

    def visit_system_tree(self, visitor):
        """Walk every node in the system tree and run a `visitor` function on
        each item.
        """
        self.system_tree.visit(visitor)

    def visit_device_tree(self, visitor):
        """Walk every node in the device tree and run a `visitor` function on
        each item.
        """
        self.device_tree.visit(visitor)

    # ---------------------------------------------------------------------
    # Traits Handlers

    def _get_online(self):
        """The topology is considered online if `_system_hash` is not empty.
        """
        return not self._system_hash.empty()

    def _system_tree_default(self):
        """We must provide a default initializer since the singleton system
        will instantiate the once instance of this object.
        """
        return SystemTree()

    def _device_tree_default(self):
        """We must provide a default initializer since the singleton system
        will instantiate the once instance of this object.
        """
        return DeviceSystemTree()

    @on_trait_change('_project_devices:device_id')
    def _project_device_instance_id_changed(self, obj, name, old, new):
        """React to the device id of a project device changing.
        """
        msg = ('ProjectDeviceInstance changed its instance_id to a name which '
               'is already used by another device!')
        assert new not in self._project_devices, msg

        proj_device = self._project_devices.pop(old, None)
        if proj_device is not None:
            self._project_devices[new] = proj_device

    # ---------------------------------------------------------------------
    # Handlers for GUI Server Messages

    def class_schema_updated(self, server_id, class_id, schema):
        """Called when a `classSchema` message is received from the server.
        """
        if len(schema.hash) > 0:
            # if it's a valid schema we cache it even if it is not requested,
            # may be useful in future
            self._class_schemas.setdefault(server_id, {})[class_id] = schema

        key = (server_id, class_id)
        if (key not in self._class_proxies.keys() and
                key not in self._project_device_proxies.keys()):
            # This schema has nothing to do with us whatsoever
            # We used to print 'Unrequested schema for classId {} arrived'.
            return

        self._requested_classes.setdefault(server_id, set()).discard(class_id)

        proxies = []
        if key not in self._class_proxies.keys():
            # We lazy build a DeviceClassProxy for future usage
            binding = build_binding(schema)
            proxy = DeviceClassProxy(server_id=server_id, binding=binding)
            self._class_proxies[key] = proxy
            apply_default_configuration(proxy.binding)
            proxy.config_update = True
        else:
            # Make sure we update our class proxies for the topology!
            proxies = [self._class_proxies[key]]

        proxies.extend(self._project_device_proxies.get(key, {}).values())
        for proxy in proxies:
            # 1. If the class schema has already arrived in the past, then we
            # should take no further action. We don't expect the schema of a
            # class on a running server to change. Device _instances_, however,
            # will change, but anyhow don't arrive in this handler.
            # 2. Only proxies are updated where we requested this!
            if (len(proxy.binding.value) > 0
                    or proxy.status != ProxyStatus.REQUESTED):
                continue
            build_binding(schema, existing=proxy.binding)
            if proxy is self._class_proxies.get(key, None):
                # Only apply default config if not a project device!
                apply_default_configuration(proxy.binding)
                proxy.config_update = True

    def device_config_updated(self, device_id, config):
        """Called when a `deviceConfiguration` message is received from the
        server

        Returns the ``DeviceProxy`` object associated with `device_id`, if
        one exists and is initialized.
        """
        proxy = self._device_proxies.get(device_id, None)
        if proxy is None or len(proxy.binding.value) == 0:
            return None

        # Apply the configuration to the proxy.
        apply_configuration(config, proxy.binding)
        # Fire the config_update event
        # XXX: Still needed with proxy.binding.config_update?
        proxy.config_update = True

        return proxy

    def device_schema_updated(self, device_id, schema):
        """Called when a `deviceSchema` message is received from the server

        Returns the ``DeviceProxy`` object associated with `schema`, if one
        exists and is initialized.
        """
        proxy = self._device_proxies.get(device_id)
        if proxy is None:
            # We used to print 'Unrequested schema for device {} arrived' here
            return None

        # Rebuild the binding object
        build_binding(schema, existing=proxy.binding)

        return proxy

    def initialize(self, server_hash):
        """Initialize the system topology with the system hash"""
        self._system_hash = server_hash
        self.system_tree.initialize(server_hash)
        self.device_tree.initialize(server_hash)

    # ---------------------------------------------------------------------
    # Topology interface

    def topology_update(self, changes):
        """Handle the bulk system topology update from the server

        This function will return the devices and servers which have been
        removed from the topology via ``slotInstanceGone``
        """
        devices, servers = [], []

        gone = changes['gone']
        if not gone.empty():
            devices, servers = self.instance_gone(gone)
        new = changes['new']
        if not new.empty():
            self.instance_new(new)
        update = changes['update']
        if not update.empty():
            self.instance_updated(update)

        return devices, servers

    def instance_new(self, server_hash):
        """Called when an `instanceNew` message is received from the server
        """
        existing_devices = self.system_tree.clear_existing(server_hash)
        # Update system topology with new configuration

        self._system_hash.merge(server_hash, "merge")

        # Update the high-level representation (Trees)
        new_topology_nodes = self.system_tree.update(server_hash)
        self.device_tree.update(server_hash)

        # Attach topology nodes and set the status
        for deviceId, node in new_topology_nodes.items():
            proxy = self._device_proxies.get(deviceId)
            if proxy is not None:
                proxy.topology_node = node
                # Set the fundamental info of the proxy
                proxy.server_id = node.attributes["serverId"]
                proxy.binding.class_id = node.attributes["classId"]
                proxy.status = ProxyStatus.ONLINE

        self._request_server_classes(server_hash)

        return existing_devices

    def instance_updated(self, server_hash):
        """Called when an `instanceUpdated` message is received from the server
        """
        self._system_hash.merge(server_hash, "merge")

        # Only the system tree takes an instance update at the moment
        # for the instanceInfo
        self.system_tree.instance_update(server_hash)
        # And the device tree!
        self.device_tree.instance_update(server_hash)

        self._request_server_classes(server_hash)

    def instance_gone(self, system_hash):
        """Remove servers, devices and macros from the system topology

        This function will return the devices and servers which have been
        removed from the topology in separate lists ``devices`` and ``servers``
        """
        devices, servers = [], []

        if 'device' in system_hash:
            self._topology_device_gone('device', system_hash, devices)
        if 'macro' in system_hash:
            self._topology_device_gone('macro', system_hash, devices)
        if 'server' in system_hash:
            self._topology_server_gone(system_hash, servers)
        if 'client' in system_hash:
            self._topology_client_gone(system_hash)

        return devices, servers

    def _topology_server_gone(self, system_hash, servers):
        """Servers were gone from the systemTopology

        :param servers: List of servers to be extended which are removed
                        from the topology
        """
        for instance_id in system_hash['server'].keys():
            path = 'server' + '.' + instance_id
            attributes = self.get_attributes(path)
            if attributes is None:
                continue

            # Remove instance from system hash
            del self._system_hash[path]

            self.system_tree.remove_server(instance_id)
            for key in self._class_proxies.keys():
                if key[0] == instance_id:
                    # Clear the class proxy binding values when device
                    # server goes offline
                    proxy = self._class_proxies[key]
                    proxy.binding.value.clear_namespace()

            self._class_schemas.setdefault(instance_id, {}).clear()
            # And erase all potential requested class schemas
            self._requested_classes.pop(instance_id, None)
            # Update status of all offline project device proxies
            self._project_device_proxies_server_update(instance_id)

            # Note the details of what device is gone
            host = attributes['host']
            servers.append((instance_id, host, ProxyStatus.OFFLINE))

    def _topology_device_gone(self, instance_type, system_hash, devices):
        """Check if devices or macros are gone in the system topology

        :param devices: List of devices and macros to be extended which are
                        removed from the topology
        """
        for instance_id, _, attr in system_hash[instance_type].iterall():
            # Get the attributes of the instance which is now gone
            path = instance_type + '.' + instance_id
            attributes = self.get_attributes(path)
            if attributes is None:
                continue

            # Remove instance from system hash
            del self._system_hash[path]

            # Set the DeviceProxy to offline, but keep it.
            proxy = self._device_proxies.get(instance_id)
            if proxy is not None:
                proxy.status = ProxyStatus.OFFLINE

            # Update the trees
            self.system_tree.remove_device(instance_id)
            self.device_tree.remove_device(instance_id)

            # Note the details of what device is gone
            class_id = attributes['classId']
            devices.append((instance_id, class_id, ProxyStatus.OFFLINE))

    def _topology_client_gone(self, system_hash):
        for instance_id, _, attr in system_hash['client'].iterall():
            path = 'client' + '.' + instance_id
            attrs = self.get_attributes(path)
            if attrs is None:
                continue
            del self._system_hash[path]

    # ---------------------------------------------------------------------
    # Utilities

    def _request_server_classes(self, server_hash):
        """Topology changed so send new class schema requests"""
        srv_hsh = server_hash.get('server', None)
        if srv_hsh is None:
            return

        for server_id in srv_hsh:
            for srv_id, cls_id in self._class_proxies.keys():
                if (srv_id == server_id and cls_id
                        in srv_hsh[srv_id, ...].get('deviceClasses', ())):
                    proxy = self._class_proxies[(srv_id, cls_id)]
                    if len(proxy.binding.value) == 0:
                        proxy.refresh_schema()

            # Update all offline devices (project device proxies)
            self._project_device_proxies_server_update(server_id)

    def _get_device_attributes(self, device_id):
        """Get the attributes for a device, if it's online. Otherwise get None.
        """
        for parent in ('device', 'macro'):
            attrs = self.get_attributes(parent + '.' + device_id)
            if attrs is not None:
                break
        return attrs

    def _project_device_proxies_server_update(self, server_id):
        """Because the server online status has changed , update the project
        device proxy's status accordingly
        """
        for (srv_id, cls_id), mapping in self._project_device_proxies.items():
            # key -> (server_id, class_id)
            # mapping -> {deviceId: project_device_proxy}
            if server_id == srv_id:
                for dev_id, proxy in mapping.items():
                    proxy.update_status()



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on January 12, 2017
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import re
from contextlib import contextmanager

from traits.api import (
    Bool, Dict, Enum, Event, HasStrictTraits, Instance, Int, List, String,
    WeakRef)

from karabo.common.api import InstanceStatus
from karabogui.itemtypes import NavigationItemTypes

HOST_LEVEL = 0
SERVER_LEVEL = 1
CLASS_LEVEL = 2
DEVICE_LEVEL = 3


class SystemTreeNode(HasStrictTraits):
    """A node in the SystemTree
    """
    node_id = String
    path = String
    status = Enum(*InstanceStatus)
    capabilities = Int
    attributes = Dict
    # Struct to keep track of all alarms related to this
    monitoring = Bool(False)
    level = Int(-1)

    parent = WeakRef('SystemTreeNode')
    children = List(Instance('SystemTreeNode'))

    def child(self, node_id):
        for child in self.children:
            if child.node_id == node_id:
                return child
        return None

    def info(self):
        level = self.level
        if level == HOST_LEVEL:
            return {'type': NavigationItemTypes.HOST,
                    "hostId": self.node_id}
        elif level == SERVER_LEVEL:
            return {'type': NavigationItemTypes.SERVER,
                    'serverId': self.node_id,
                    'log': self.attributes.get("log", "INFO"),
                    'karaboVersion': self.attributes.get('karaboVersion',
                                                         "0.0.0")}
        elif level == CLASS_LEVEL:
            return {'type': NavigationItemTypes.CLASS,
                    'serverId': self.parent.node_id,
                    'classId': self.node_id,
                    'deviceId': ''}
        elif level == DEVICE_LEVEL:
            return {'type': NavigationItemTypes.DEVICE,
                    'classId': self.parent.node_id,
                    'deviceId': self.node_id,
                    'archive': self.attributes.get("archive", False),
                    'capabilities': self.capabilities}

    def row(self):
        if self.parent is None:
            return 0
        return self.parent.children.index(self)


class SystemTree(HasStrictTraits):
    """A data model which holds data concerning the devices and servers in a
    running Karabo system.
    """
    # Base node for the whole tree
    root = Instance(SystemTreeNode, args=())

    # A context manager to enter when manipulating the tree
    update_context = Instance(object)

    # An event which is triggered whenenver the tree needs a layout update
    status_update = Event

    # device/server node lookup dicts
    _device_nodes = Dict
    _server_nodes = Dict

    # initialized hook
    initialized = Event

    def clear_all(self):
        """Removes all data from the model.
        """
        self._device_nodes.clear()
        self._server_nodes.clear()

        with self.update_context.reset_context():
            self.root.children = []

    def clear_existing(self, system_hash):
        """Checks whether instances already exist in the tree and removes them.

        Returns a list with all existing device_ids and a list with existing
        server class keys (server_id, class_id).
        """
        existing_devices = []
        server_hash = system_hash.get('server', {})
        device_hash = system_hash.get('device', {})

        for server_id in server_hash.keys():
            server_classes = self.remove_server(server_id)
            # Check, if server_id was in tree
            if server_classes:
                existing_devices.append(server_id)

        for device_id in device_hash.keys():
            # device_id might not be in the tree
            if not self.remove_device(device_id):
                continue
            existing_devices.append(device_id)

        return existing_devices

    def find(self, node_id, case_sensitive=True,
             use_reg_ex=True, full_match=False):
        """Find all nodes with the given `node_id` and return them in a list

        :param node_id: The actual string we are looking for in the tree
        :param case_sensitive: States whether the given string should match
                               case or not
        :param full_match: States whether the given string matches fully
        :param use_reg_ex: Defines the given string as a regular expression
        :return list of found nodes (which could be empty)
        """

        found_nodes = []
        pattern = node_id if use_reg_ex else f".*{re.escape(node_id)}"
        flags = 0 if case_sensitive else re.IGNORECASE
        regex = re.compile(pattern, flags=flags)

        def visitor(node):
            parent_node = node.parent
            parent_check = parent_node is None
            if parent_check:
                # Do not look for nodes which are its parent
                return

            matcher = regex.fullmatch if full_match else regex.match
            if matcher(node.node_id) is not None:
                found_nodes.append(node)

        self.visit(visitor)
        return found_nodes

    def get_instance_node(self, instance_id):
        """Retrieve a tree node for a device or server instance"""
        if instance_id in self._device_nodes:
            return self._device_nodes[instance_id]
        if instance_id in self._server_nodes:
            return self._server_nodes[instance_id]
        return None

    def remove_device(self, instance_id):
        """Remove the entry for a device from the tree
        """
        node = self._device_nodes.pop(instance_id, None)
        if node is None:
            return False

        class_node = node.parent
        with self.update_context.removal_context(node):
            class_node.children.remove(node)

        if not class_node.children:
            server_node = class_node.parent
            with self.update_context.removal_context(class_node):
                server_node.children.remove(class_node)

        return True

    def remove_server(self, instance_id):
        """Remove the entry for a server from the tree
        """
        server_node = self._server_nodes.pop(instance_id, None)
        server_class_keys = []

        if server_node is None:
            return server_class_keys

        # Take care of removing all children from bottom to top
        for class_node in server_node.children:
            for device_node in class_node.children:
                self._device_nodes.pop(device_node.node_id, None)
            with self.update_context.removal_children_context(class_node):
                class_node.children = []

        with self.update_context.removal_children_context(server_node):
            # Remove class nodes
            server_node.children = []

        # Finally remove server node
        host_node = server_node.parent
        with self.update_context.removal_context(server_node):
            host_node.children.remove(server_node)

        if not host_node.children:
            self._remove_root_children(host_node)

        return server_class_keys

    def update(self, system_hash):
        """This function is called whenever the system topology has changed
        and the view needs an update.

        The incoming ``config`` represents the system topology.

        Returns a dictionary of ``SystemTreeNode`` instances for newly added
        device instances.
        """
        with self.update_context.layout_context():
            self._handle_server_data(system_hash)
            nodes = self._handle_device_data('device', system_hash)
            nodes.update(self._handle_device_data('macro', system_hash))

        return nodes

    def instance_update(self, system_hash):
        """Put the contents of Hash `system_hash` into the internal tree
        structure.
        """

        device_status = set()

        if 'server' in system_hash:
            for server_id, _, attrs in system_hash['server'].iterall():
                if len(attrs) == 0:
                    continue
                server_node = self._server_nodes.get(server_id, None)
                if server_node is None:
                    continue
                server_node.attributes = attrs

        for device_type in ('device', 'macro'):
            if device_type in system_hash:
                for device_id, _, attrs in system_hash[device_type].iterall():
                    if len(attrs) == 0:
                        continue

                    device_node = self._device_nodes.get(device_id, None)
                    if device_node is None:
                        continue

                    device_status.add(device_id)
                    status = InstanceStatus(attrs['status'])
                    device_node.status = status
                    device_node.attributes = attrs

        if device_status:
            self.status_update = device_status

    def initialize(self, system_hash):
        """Initialize the system topology with a reset context
        """
        with self.update_context.reset_context():
            self._handle_server_data(system_hash, append=False)
            self._handle_device_data('device', system_hash, append=False)
            self._handle_device_data('macro', system_hash, append=False)
        self.initialized = True

    def visit(self, visitor):
        """Walk every node of the system tree until visitor returns true

        Call the `visitor` function on each item, if the function returns
        a true-ish value, the loop will be stopped.
        """

        def _iter_tree_node(node):
            yield node
            for child in node.children:
                yield from _iter_tree_node(child)

        for t_node in _iter_tree_node(self.root):
            if visitor(t_node):
                break

    # ------------------------------------------------------------------

    def _update_context_default(self):
        """Traits default initializer for the `update_context` trait.
        """

        class Dummy:
            @contextmanager
            def reset_context(self):
                yield

            @contextmanager
            def insertion_context(self, parent_node, first, last):
                yield

            @contextmanager
            def removal_context(self, tree_node):
                yield

            @contextmanager
            def layout_context(self):
                yield

            @contextmanager
            def insert_root_context(self, first, last):
                yield

            @contextmanager
            def remove_root_context(self, first, last):
                yield

            @contextmanager
            def removal_children_context(self, tree_node):
                yield

        return Dummy()

    def _append_child_node(self, parent_node, child_node):
        # First and last indices, INCLUSIVE
        first = len(parent_node.children)
        last = first

        if child_node.level == SERVER_LEVEL:
            self._server_nodes[child_node.node_id] = child_node
        elif child_node.level == DEVICE_LEVEL:
            self._device_nodes[child_node.node_id] = child_node

        with self.update_context.insertion_context(parent_node, first, last):
            parent_node.children.append(child_node)

    def _set_child_node(self, parent_node, child_node):
        """Set the child node directly without announcing layout change

        NOTE: This function can be only used in a reset context!
        """
        if child_node.level == SERVER_LEVEL:
            self._server_nodes[child_node.node_id] = child_node
        elif child_node.level == DEVICE_LEVEL:
            self._device_nodes[child_node.node_id] = child_node

        parent_node.children.append(child_node)

    def _set_root_children(self, host_node, append=True):
        """Set the a host node in the root element"""
        if append:
            first = len(self.root.children)
            last = first
            with self.update_context.insert_root_context(first, last):
                self.root.children.append(host_node)
        else:
            self.root.children.append(host_node)

    def _remove_root_children(self, node):
        """Remove a node in the root element"""
        first = node.row()
        last = first
        with self.update_context.remove_root_context(first, last):
            self.root.children.remove(node)

    # ------------------------------------------------------------------

    def _handle_server_data(self, system_hash, append=True):
        """Put the contents of Hash `system_hash` into the internal tree
        structure.
        """
        if 'server' not in system_hash:
            return

        handler = self._append_child_node if append else self._set_child_node
        for server_id, _, attrs in system_hash['server'].iterall():
            if len(attrs) == 0:
                continue

            host = attrs['host']

            # Create node for host
            host_node = self.root.child(host)
            if host_node is None:
                host_node = SystemTreeNode(node_id=host, path=host,
                                           parent=self.root,
                                           level=HOST_LEVEL)
                self._set_root_children(host_node, append)

            # Create node for server
            server_node = host_node.child(server_id)
            if server_node is None:
                server_node = SystemTreeNode(node_id=server_id,
                                             path=server_id, parent=host_node,
                                             level=SERVER_LEVEL)
                handler(host_node, server_node)
            server_node.attributes = attrs

    def _handle_device_data(self, device_type, system_hash, append=True):
        """Put the contents of Hash `system_hash` into the internal tree
        structure. Returns a dictionary of any newly added device
        ``SystemTreeNode`` instances.
        """
        new_dev_nodes = {}
        assert device_type in ('device', 'macro')

        if device_type not in system_hash:
            return new_dev_nodes

        handler = self._append_child_node if append else self._set_child_node
        for device_id, _, attrs in system_hash[device_type].iterall():
            if len(attrs) == 0:
                continue

            host = attrs['host']
            capabilities = attrs['capabilities']
            server_id = attrs['serverId']
            class_id = attrs['classId']
            status = InstanceStatus(attrs['status'])

            # A device must have a server added before with host!
            host_node = self.root.child(host)
            if host_node is None:
                continue

            server_node = host_node.child(server_id)
            if server_node is None:
                continue

            # Class node
            class_node = server_node.child(class_id)
            if class_node is None:
                server_id = server_node.node_id
                path = f'{server_id}.{class_id}'
                class_node = SystemTreeNode(node_id=class_id,
                                            path=path, parent=server_node,
                                            level=CLASS_LEVEL)
                handler(server_node, class_node)

            # Device node
            device_node = class_node.child(device_id)
            if device_node is None:
                device_node = SystemTreeNode(node_id=device_id, path=device_id,
                                             parent=class_node,
                                             status=status,
                                             capabilities=capabilities,
                                             attributes=attrs,
                                             level=DEVICE_LEVEL)
                handler(class_node, device_node)
                # new nodes should be returned
                new_dev_nodes[device_id] = device_node

        return new_dev_nodes



##############################################################################
# Created on September 13, 2022
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################

import getpass
import json
from datetime import timedelta
from enum import IntEnum, unique
from struct import calcsize, unpack

from qtpy import uic
from qtpy.QtCore import Qt, QTimer, QUrl, Signal, Slot
from qtpy.QtGui import QDesktopServices, QIntValidator, QKeyEvent, QKeySequence
from qtpy.QtNetwork import (
    QAbstractSocket, QNetworkAccessManager, QNetworkReply, QNetworkRequest,
    QSslConfiguration, QSslSocket, QTcpSocket)
from qtpy.QtWidgets import (
    QApplication, QDialog, QDialogButtonBox, QHBoxLayout, QLineEdit,
    QSizePolicy, QWidget)

from karabo.native import decodeBinary
from karabogui import access as krb_access
from karabogui.const import CLIENT_HOST
from karabogui.events import (
    KaraboEvent, register_for_broadcasts, unregister_from_broadcasts)
from karabogui.singletons.api import get_config, get_network
from karabogui.util import SignalBlocker, utc_to_local_date

from .utils import get_dialog_ui

TIMER_DELAY = 500  # ms
TOKEN_CHECK_TIME = 3000  # 3s
REQUEST_HEADER = "application/json"
CONNECTION_CHECK_TIME = 5000  # 5s

USER_INFO = """<p>You are logged in as '<span style=" font-weight:600;">
{username}</span>'. Click 'Connect' to continue or</p><p>'Switch User' to
change the user.</p>"""

ACCESS_LEVEL_INFO = """<html><head/><body><p><span style=" font-size:10pt;
font-style:italic; color:#242424;">Currently logged in as <b>'{user}'</b>
with the access level <b>'{access_level}'.</b></span></p><p><span style="
font-size:10pt; font-style:italic; color:#242424;">Open Access Form to
generate the access code by authenticating the user.
</span></p></body></html>"""

NO_REFRESH_TOKEN_ERROR = "Refresh Token not found"


@unique
class LoginType(IntEnum):
    UNKNOWN = 0
    USER_AUTHENTICATED = 1
    ACCESS_LEVEL = 2
    READ_ONLY = 3
    REFRESH_TOKEN = 4


class Validator(QIntValidator):
    """A custom validator that allows only 1 or 6 digits """

    def validate(self, input, pos):
        """Support only number with single digit or 6 digits"""
        if input in ("+", "-"):
            return self.Invalid, input, pos
        if not (pos and input):
            # On deleting item from cell
            return self.Intermediate, input, pos
        if len(input) not in (1, 6):
            return self.Invalid, input, pos
        if pos not in (1, 6):
            return self.Invalid, input, pos
        return super().validate(input, pos)


class Cell(QLineEdit):
    onBackspacePressed = Signal()
    moveToNextCell = Signal(bool)
    onPasteNewCode = Signal()

    def keyPressEvent(self, event: QKeyEvent) -> None:
        if event.key() == Qt.Key_Backspace:
            self.onBackspacePressed.emit()
        elif event.key() == Qt.Key_Right:
            self.moveToNextCell.emit(True)
        elif event.key() == Qt.Key_Left:
            self.moveToNextCell.emit(False)
        elif event.matches(QKeySequence.Paste):
            self.onPasteNewCode.emit()
        return super().keyPressEvent(event)


class AccessCodeWidget(QWidget):
    """A widget to display each digit in the access code in a separate cell(
    QLineEdit)"""

    valueChanged = Signal()

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setObjectName("AccessCodeWidget")

        self.cells = []
        for i in range(6):
            object_name = f"Cell_{i}"
            cell = Cell(parent=self)
            cell.setObjectName(object_name)
            cell.setFixedWidth(30)
            cell.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
            cell.setAlignment(Qt.AlignCenter)
            cell.setValidator(Validator(parent=self))
            self.cells.append(cell)
            layout.addWidget(cell)
            cell.textChanged.connect(self.on_text_changed)
            cell.onBackspacePressed.connect(self.on_backspace_pressed)
            cell.moveToNextCell.connect(self._on_move_to_cell)
            cell.onPasteNewCode.connect(self._paste_new_code)

    @Slot(str)
    def on_text_changed(self, text: str):
        current_index = self.cells.index(self.sender())
        # Distribute each digits in each cells.
        if len(text) == 6:
            self._paste_on_all_cell(text)
        # Paste one digit to current cell and move focus to the next.
        elif current_index < 5 and self.sender().text():
            next_index = current_index + 1
            self.cells[next_index].setFocus()
        self.valueChanged.emit()

    @Slot()
    def on_backspace_pressed(self) -> None:
        """Set focus on the previous cell when the cell is emtpy or the
        cursor is left to the digit."""
        sender = self.sender()
        index = self.cells.index(sender)
        if index > 0 and not (sender.text() and sender.cursorPosition()):
            index -= 1
        self.cells[index].setFocus()

    @Slot(bool)
    def _on_move_to_cell(self, forward: bool) -> None:
        """Move the cursor to the next/previous cell"""
        index = self.cells.index(self.sender())
        if forward and index != 5:
            index += 1
        if not forward and index != 0:
            index -= 1
        self.cells[index].setFocus()

    @Slot()
    def _paste_new_code(self):
        text = QApplication.clipboard().text()
        if len(text) == 6 and text.isdigit():
            self._paste_on_all_cell(text)
            self.valueChanged.emit()

    def get_access_code(self) -> str:
        return "".join(cell.text() for cell in self.cells)

    def has_access_code(self):
        return all([cell.text() for cell in self.cells])

    def _paste_on_all_cell(self, text):
        with SignalBlocker(self):
            for number, cell in zip(str(text), self.cells):
                cell.setText(number)


BUTTON_STYLE = """
QPushButton#skip_authentication_button{border: none; color: rgb(80,133,207)}
QPushButton#skip_authentication_button:hover{border: none;
                                             color: rgb(8,8,245)}"""


class ReactiveLoginDialog(QDialog):
    def __init__(self, access_level="expert", hostname="", port="",
                 gui_servers=None, parent=None):
        super().__init__(parent)
        filepath = get_dialog_ui("reactive_login_dialog.ui")
        uic.loadUi(filepath, self)
        if parent is None:
            self.setWindowFlags(self.windowFlags() | Qt.WindowStaysOnTopHint)

        self.stackedWidget.setCurrentIndex(0)
        self.login_type = LoginType.UNKNOWN
        self._switch_user_pressed = False
        self._requesting = False
        self._queried_host = ""
        self._queried_port = ""
        self._authenticating = False
        self._auth_url = ""
        self._auth_user = ""
        self._error = None
        self._topic = ""

        # Initialization of communication objects needed for authenticating
        # users.
        self.access_manager = QNetworkAccessManager()
        self.access_manager.finished.connect(self.onAuthReply)
        # XXX: Remove the code below that disables SSL certification path
        # verification as soon as real production SSL certificates are
        # used.
        ssl = QSslConfiguration.defaultConfiguration()
        ssl.setPeerVerifyMode(QSslSocket.VerifyNone)
        QSslConfiguration.setDefaultConfiguration(ssl)

        # The socket used to get information from a GUI Server
        self._tcp_socket = QTcpSocket(self)
        self._tcp_socket.readyRead.connect(self.onReadServerData)
        self._tcp_socket.error.connect(self.onSocketError)

        self._timer = QTimer(self)
        self._timer.setInterval(TIMER_DELAY)
        self._timer.setSingleShot(True)
        self._timer.timeout.connect(self.connectToServer)

        self._connection_check_timer = QTimer(self)
        self._connection_check_timer.setInterval(CONNECTION_CHECK_TIME)
        self._connection_check_timer.timeout.connect(self._connect_to_server)
        self._connection_check_timer.start()

        self._token_check_timer = QTimer(parent=self)
        self._token_check_timer.timeout.connect(self._look_for_token)

        hostname = hostname if hostname else "localhost"
        if gui_servers is None:
            gui_servers = []
        for server in gui_servers:
            self.combo_hostname.addItem(server)
        self.combo_hostname.setEditText(hostname)
        self.combo_hostname.editTextChanged.connect(self.on_hostname_changed)

        port = str(port) if port else "44444"
        self.edit_port.setText(port)
        self.edit_port.setValidator(QIntValidator(1, 65535, parent=self))
        self.edit_port.textEdited.connect(self.on_port_changed)

        for level in ["expert", "operator", "observer"]:
            self.combo_access_level.addItem(level)
        access = self.combo_access_level.findText(access_level)
        self.combo_access_level.setCurrentIndex(access)

        # Access Level and Read-Only login
        shell_user = getpass.getuser()
        self.label_shell_user.setText(shell_user)
        self.label_readonly_user.setText(shell_user)

        self.edit_access_code = AccessCodeWidget(parent)
        self.access_widget_layout.addWidget(self.edit_access_code)
        self.edit_access_code.valueChanged.connect(self._update_button)

        self.authenticate_button.clicked.connect(self.open_login_webpage)

        self.connect_button.clicked.connect(self.connect_clicked)
        self._update_dialog_state()

        if self.hasAcceptableInput():
            self._timer.start()

        self.switch_button.clicked.connect(self._on_switch_user_click)
        self.skip_authentication_button.setStyleSheet(BUTTON_STYLE)
        self.skip_authentication_button.clicked.connect(self.accept)

    # --------------------------------------------------------------------
    # Dialog Public Properties

    @property
    def access_level(self) -> str:
        level = get_config()["access_level"]
        if self.login_type is LoginType.ACCESS_LEVEL:
            level = self.combo_access_level.currentText()
        return level

    @property
    def hostname(self):
        return self.combo_hostname.currentText()

    @property
    def port(self):
        return int(self.edit_port.text())

    @property
    def gui_servers(self):
        return [self.combo_hostname.itemText(i)
                for i in range(self.combo_hostname.count())]

    def hasAcceptableInput(self):
        """Return if the dialog has acceptable input for the host and port"""
        host = len(self.combo_hostname.currentText()) > 3
        port = len(self.edit_port.text()) > 0
        return host and port

    # --------------------------------------------------------------------
    # Qt Slots

    @Slot(str)
    def on_hostname_changed(self, value):
        """Split the selected text into hostname and port and trigger the
        connect to server timer.
        """
        if ":" in value:
            hostname, port = value.split(":")
            self.combo_hostname.setEditText(hostname)
            if port.isnumeric():
                self.edit_port.setText(port)

        if self.hasAcceptableInput():
            self._timer.start()

    @Slot()
    def on_port_changed(self):
        if self.hasAcceptableInput():
            self._timer.start()
        self._update_button()

    @Slot()
    def connect_clicked(self):
        self._error = None
        if self.login_type is LoginType.USER_AUTHENTICATED:
            self._login_authenticated()
        elif self.login_type is LoginType.REFRESH_TOKEN:
            self._refresh_authentication()
        else:
            self.accept()

    @Slot()
    def connectToServer(self):
        """Connect to the server and update the dialog."""
        self._connect_to_server()
        self._update_dialog_state()

    @Slot()
    def _connect_to_server(self):
        """Connect to the server"""
        if self._tcp_socket.state() != QTcpSocket.UnconnectedState:
            self._tcp_socket.abort()

        self._queried_port = self.edit_port.text()
        if not self._queried_port:
            return
        self._requesting = True
        self._queried_host = self.combo_hostname.currentText()
        self._topic = ""
        self._error = None
        self.login_type = LoginType.UNKNOWN
        self._tcp_socket.connectToHost(self._queried_host,
                                       int(self._queried_port))

    @Slot()
    def onReadServerData(self):
        """Read data coming via the tcp socket"""
        bytesNeededSize = calcsize("I")
        rawBytesNeeded = self._tcp_socket.read(bytesNeededSize)
        bytesNeeded = unpack("I", rawBytesNeeded)[0]

        dataBytes = self._tcp_socket.read(bytesNeeded)
        server_info = decodeBinary(dataBytes)

        if not self._requesting:
            # The connection info received is not for the latest attempt to
            # connect to a GUI Server - ignore it.
            self._tcp_socket.disconnectFromHost()
            self._update_dialog_state()
            return

        if server_info.has("readOnly") and server_info["readOnly"]:
            self.login_type = LoginType.READ_ONLY
        elif (server_info.has("authServer") and
              bool(server_info["authServer"])):
            self.login_type = LoginType.USER_AUTHENTICATED
            if (not self._switch_user_pressed
                    and get_config()["refresh_token"] is not None):
                # If the user has already pressed "Switch User"
                # we don't go back to REFRESH_TOKEN mode
                self.login_type = LoginType.REFRESH_TOKEN
            self._auth_url = server_info["authServer"]
            if not self._auth_url.endswith("/"):
                self._auth_url += "/"
            krb_access.AUTHENTICATION_SERVER = self._auth_url
        else:
            self.login_type = LoginType.ACCESS_LEVEL

        if self.login_type in (LoginType.USER_AUTHENTICATED,
                               LoginType.REFRESH_TOKEN):
            self._token_check_timer.start(TOKEN_CHECK_TIME)
        elif self._token_check_timer.isActive():
            self._token_check_timer.stop()

        self._topic = f"Topic: {server_info['topic']}"
        self._tcp_socket.disconnectFromHost()
        self._requesting = False
        self._update_dialog_state()

    @Slot(QAbstractSocket.SocketError)
    def onSocketError(self, error):
        self._requesting = False
        self._error = ("No GUI Server available at "
                       f"{self._queried_host}:{self._queried_port}")
        self._update_dialog_state()
        if self._token_check_timer.isActive():
            self._token_check_timer.stop()

    @Slot(QNetworkReply)
    def onAuthReply(self, reply):
        self._authenticating = False
        self._update_dialog_state()

        url = reply.url().path()
        if url.endswith("user_tokens") and not self.remember_login.isChecked():
            self._clear_refresh_token()

        error = reply.error()
        bytes_string = reply.readAll()
        reply_body = str(bytes_string, "utf-8")
        auth_result = json.loads(reply_body)
        if error == QNetworkReply.NoError and auth_result["success"]:
            krb_access.ONE_TIME_TOKEN = auth_result["once_token"]
            refresh_token = auth_result.get("refresh_token")
            if refresh_token is not None:
                refresh_token_user = auth_result.get("username")
                get_config()["refresh_token"] = refresh_token
                get_config()["refresh_token_user"] = refresh_token_user
            self.accept()
        else:
            self._error = auth_result.get("error_msg")
            if self._error == NO_REFRESH_TOKEN_ERROR:
                self._clear_refresh_token()
            self.stackedWidget.setCurrentIndex(LoginType.USER_AUTHENTICATED)
            self._update_status_label()

        reply.deleteLater()

    @Slot()
    def open_login_webpage(self):
        """Open browser to load the login page."""
        url = f"{self._auth_url}login_form"
        success = QDesktopServices.openUrl(QUrl(url))
        if not success:
            self._error = f"Failed to open the page {url}."
            self._update_status_label()

    # --------------------------------------------------------------------
    # Private interface

    def _update_dialog_state(self):
        """Update the stacked widget of the dialog and the status label"""
        self.label_topic.setText(self._topic)
        self.label_topic.setVisible(bool(self._topic))
        self.stackedWidget.setCurrentIndex(self.login_type.value)

        self._update_status_label()
        # Status of Connect button
        self._update_button()

        if self.login_type is LoginType.REFRESH_TOKEN:
            text = USER_INFO.format(
                username=get_config()["refresh_token_user"])
            self.user_info_label.setText(text)

    def _update_button(self):
        """Update the connect button according to the status"""
        if self.login_type in (LoginType.ACCESS_LEVEL, LoginType.READ_ONLY):
            self.connect_button.setEnabled(True)
        elif self.login_type is LoginType.UNKNOWN:
            self.connect_button.setEnabled(False)
        elif self.login_type is LoginType.USER_AUTHENTICATED:
            enable = self.edit_access_code.has_access_code()
            self.connect_button.setEnabled(enable)
        elif self.login_type is LoginType.REFRESH_TOKEN:
            self.connect_button.setEnabled(True)

    def _update_status_label(self):
        """Status message and processing indicator"""
        status_color = "black"
        status_text = ""
        if self._authenticating:
            status_text = (
                f"<b>Authenticating</b> user <b>{self._auth_user}</b> ...")
        elif self._requesting:
            status_text = (
                "Getting <b>info</b> for Server "
                f"<b>{self._queried_host}:{self._queried_port}</b> ...")
        elif self._error is not None:
            status_color = "red"
            status_text = self._error
        self.label_status.setStyleSheet(f"QLabel {{color: {status_color};}}")
        self.label_status.setText(status_text)

    def _login_authenticated(self):
        """Authentication Request Posting"""
        self._authenticating = True
        self._update_dialog_state()

        remember_login = self.remember_login.isChecked()
        info = json.dumps({
            "access_code": int(self.edit_access_code.get_access_code()),
            "client_hostname": CLIENT_HOST,
            "remember_login": remember_login
        })
        if not remember_login:
            # Erase existing information
            self._clear_refresh_token()

        info = bytearray(info.encode("utf-8"))
        url = f"{self._auth_url}user_tokens"

        request = QNetworkRequest(QUrl(url))
        request.setHeader(
            QNetworkRequest.ContentTypeHeader, REQUEST_HEADER)
        self.access_manager.post(request, info)

    def _refresh_authentication(self):
        self._authenticating = True
        self._update_dialog_state()

        info = json.dumps({
            "refresh_token": get_config()["refresh_token"],
            "username": get_config()["refresh_token_user"],
            "client_hostname": CLIENT_HOST,
        })
        info = bytearray(info.encode("utf-8"))
        url = f"{self._auth_url}refresh_tokens"

        request = QNetworkRequest(QUrl(url))
        request.setHeader(QNetworkRequest.ContentTypeHeader, REQUEST_HEADER)
        self.access_manager.post(request, info)

    def _clear_refresh_token(self):
        del get_config()["refresh_token"]
        del get_config()["refresh_token_user"]

    @Slot()
    def _on_switch_user_click(self):
        self.login_type = LoginType.USER_AUTHENTICATED
        self._switch_user_pressed = True
        self._update_dialog_state()
        if self._token_check_timer.isActive():
            self._token_check_timer.stop()

    @Slot()
    def _look_for_token(self):
        if get_config()["refresh_token"] is None:
            self.login_type = LoginType.USER_AUTHENTICATED
        elif not self._switch_user_pressed:
            # Refresh token mode should not be activated once
            # the user has pressed "Switch User"
            self.login_type = LoginType.REFRESH_TOKEN
        self._update_dialog_state()


TEMPORARY_INDEX = 0
PERMANENT_INDEX = 1


class UserSessionDialog(QDialog):
    def __init__(self, parent: QWidget | None = None):
        super().__init__(parent)
        self.setModal(False)
        filepath = get_dialog_ui("user_session_dialog.ui")
        uic.loadUi(filepath, self)
        if parent is None:
            self.setWindowFlags(self.windowFlags() | Qt.WindowStaysOnTopHint)

        self.remember_login.setVisible(False)
        self.ok_button = self.buttonBox.button(QDialogButtonBox.Ok)
        self.ok_button.setEnabled(False)
        self.combo_mode.currentIndexChanged.connect(self._switch_temporary)
        if krb_access.is_end_of_session():
            self.combo_mode.setCurrentIndex(PERMANENT_INDEX)
            self.combo_mode.setEnabled(False)

        self.edit_access_code = AccessCodeWidget(parent)
        self.access_widget_layout.addWidget(self.edit_access_code)
        self.edit_access_code.valueChanged.connect(self._update_button)
        self.authenticate_button.clicked.connect(self.open_login_webpage)

        self.access_manager = QNetworkAccessManager()
        self.access_manager.finished.connect(self.onAuthReply)
        self._auth_url = krb_access.AUTHENTICATION_SERVER

        user = get_config()["username"]
        access_level = krb_access.GLOBAL_ACCESS_LEVEL.name
        text = ACCESS_LEVEL_INFO.format(user=user, access_level=access_level)
        self.info_label.setText(text)

        self.event_map = {
            KaraboEvent.NetworkConnectStatus: self._event_network,
            KaraboEvent.UserSession: self._event_user_session,
            KaraboEvent.UserSessionInfo: self._event_user_session_info,
        }
        register_for_broadcasts(self.event_map)

        get_network().onGetGuiSessionInfo()

    def done(self, result):
        """Stop listening for broadcast events"""
        unregister_from_broadcasts(self.event_map)
        super().done(result)

    # --------------------------------------------------------------------
    # Qt Slots

    @Slot(QNetworkReply)
    def onAuthReply(self, reply: QNetworkReply):
        error = reply.error()
        bytes_string = reply.readAll()
        reply_body = str(bytes_string, "utf-8")
        auth_result = json.loads(reply_body)
        if error == QNetworkReply.NoError and auth_result["success"]:
            krb_access.ONE_TIME_TOKEN = auth_result["once_token"]
            # Temporary Session
            if self.combo_mode.currentIndex() == TEMPORARY_INDEX:
                highest_access_level = krb_access.HIGHEST_ACCESS_LEVEL.name
                krb_access_level = krb_access.ACCESS_LEVEL_MAP.get(
                    highest_access_level.lower(), 2)
                username = get_config()["username"]
                get_network().beginTemporarySession(
                    username=username,
                    temporarySessionToken=auth_result["once_token"],
                    levelBeforeTemporarySession=krb_access_level,
                )
            else:
                refresh_token = auth_result.get("refresh_token")
                if refresh_token is not None:
                    refresh_token_user = auth_result.get("username")
                    get_config()["refresh_token"] = refresh_token
                    get_config()["refresh_token_user"] = refresh_token_user
                else:
                    # In case we haven't received a refresh token, we are not
                    # remembering and thus have to erase tokens.
                    self._clear_refresh_token()
                get_network().onLogin()

            super().accept()
        else:
            error = auth_result.get("error_msg")
            self.error_label.setStyleSheet(
                "QLabel#error_label {color:red;}")
            self.error_label.setText(error)
        reply.deleteLater()

    @Slot(int)
    def _switch_temporary(self, index):
        self.remember_login.setVisible(bool(index))

    @Slot()
    def open_login_webpage(self):
        """Open browser to load the login page."""
        url = f"{self._auth_url}login_form"
        success = QDesktopServices.openUrl(QUrl(url))
        if not success:
            self.error_label.setText(f"Failed to open the page {url}.")

    @Slot()
    def accept(self):
        # We are calling the super method only after the authentication of
        # the temporary session user.
        self._login_authenticated()

    @Slot()
    def _update_button(self):
        enable = self.edit_access_code.has_access_code()
        self.ok_button.setEnabled(enable)

    # --------------------------------------------------------------------
    # Private interface

    def _event_network(self, data):
        if not data.get("status"):
            self.close()

    def _event_user_session_info(self, data):
        """Work on the user session information from the gui server"""
        start_time = data["sessionStartTime"]
        duration = int(data["sessionDuration"])
        start_str, end_str, time_remaining_str = remaining_time_info(
            start_time, duration)

        self.info_start_time.setText(start_str)
        self.info_end_time.setText(end_str)
        self.info_remaining_time.setText(time_remaining_str)

    def _event_user_session(self, data):
        self.close()

    def _login_authenticated(self):
        if self.combo_mode.currentIndex() == TEMPORARY_INDEX:
            remember_login = False
        else:
            remember_login = self.remember_login.isChecked()

        info = json.dumps({
            "access_code": int(self.edit_access_code.get_access_code()),
            "client_hostname": CLIENT_HOST,
            "remember_login": remember_login,
        })
        info = bytearray(info.encode("utf-8"))
        url = f"{self._auth_url}user_tokens"

        request = QNetworkRequest(QUrl(url))
        request.setHeader(QNetworkRequest.ContentTypeHeader, REQUEST_HEADER)
        self.access_manager.post(request, info)

    def _clear_refresh_token(self):
        del get_config()["refresh_token"]
        del get_config()["refresh_token_user"]


def remaining_time_info(start: str, duration: int) -> tuple[str, str, str]:
    """Calculate the remaining information from the guiserver info

    :param start: utc aware string representation with microseconds
    :param duration: session duration in seconds (int)
    """
    dt = utc_to_local_date(start, format="%Y%m%dT%H%M%S.%fZ")
    end_local = dt + timedelta(seconds=duration)

    # Time remaining from now
    now = utc_to_local_date("")
    time_left = end_local - now

    time_left_sec = time_left.total_seconds()
    if time_left_sec <= 0:
        time_remaining_str = "Session ended"
    else:
        total_minutes = int(time_left_sec // 60)
        total_hours = float(time_left_sec / 3600)
        total_days = float(time_left_sec / 86400)

        if total_days >= 1:
            time_remaining_str = f"~{total_days:.2f} day(s) left"
        elif total_hours >= 1:
            time_remaining_str = f"~{total_hours:.2f} hour(s) left"
        else:
            time_remaining_str = f"~{total_minutes} minute(s) left"

    # Format as string
    time_format = "%Y-%m-%d %H:%M:%S"
    start_str = dt.strftime(time_format)
    end_str = end_local.strftime(time_format)

    return start_str, end_str, time_remaining_str



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import pytest
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QToolButton

from karabo.common.states import State
from karabo.native import AccessLevel, Configurable, Slot, String
from karabogui.binding.api import (
    DeviceProxy, PropertyProxy, apply_default_configuration, build_binding)
from karabogui.testing import set_proxy_value, singletons


class EmptyDevice(Configurable):
    """What a device looks like to a scene when it's offline"""


class SlottedDevice(Configurable):
    state = String(defaultValue=State.INIT)

    @Slot(allowedStates=[State.INIT],
          displayedName="Call ME",
          requiredAccessLevel=AccessLevel.OBSERVER)
    def callme(self):
        pass

    @Slot(allowedStates=[State.ACTIVE],
          requiredAccessLevel=AccessLevel.OBSERVER)
    def yep(self):
        pass


def get_slot_proxy(additional=False, klass=SlottedDevice):
    schema = klass.getClassSchema()
    binding = build_binding(schema)
    apply_default_configuration(binding)

    device = DeviceProxy(device_id="dev", server_id="Test", binding=binding)
    slot_proxy = PropertyProxy(root_proxy=device, path="callme")
    if additional:
        return slot_proxy, PropertyProxy(root_proxy=device, path="yep")
    else:
        return slot_proxy


@pytest.fixture
def icon_command_setup(gui_app):
    # import required at this point
    from ..icon_command import DisplayIconCommand

    # setup
    slot_proxy = get_slot_proxy()
    controller = DisplayIconCommand(proxy=slot_proxy)
    controller.create(None)
    yield controller
    # teardown
    controller.destroy()
    assert controller.widget is None


@pytest.fixture
def icon_command_dialog_setup(gui_app):
    from ..icon_command import IconSelectionDialog
    dialog = IconSelectionDialog()
    yield dialog
    dialog.destroy()


def test_icon_command_basics(icon_command_setup):
    """Test the basics of the display icon command"""
    controller = icon_command_setup
    assert controller.widget is not None
    assert isinstance(controller._button, QToolButton)


def test_icon_command_trigger(icon_command_setup, mocker):
    controller = icon_command_setup
    network = mocker.Mock()
    with singletons(network=network):
        controller._button.clicked.emit(True)
        network.onExecute.assert_called_with("dev", "callme", False)


def test_icon_command_state_change(gui_app):
    from ..icon_command import DisplayIconCommand
    schema = SlottedDevice.getClassSchema()
    binding = build_binding(schema)
    apply_default_configuration(binding)

    dev_proxy = DeviceProxy(device_id="dev", server_id="swerver",
                            binding=binding)
    slot_proxy = PropertyProxy(root_proxy=dev_proxy, path="yep")
    controller = DisplayIconCommand(proxy=slot_proxy)
    controller.create(None)

    # device state is INIT
    assert not controller._button.isEnabled()

    set_proxy_value(slot_proxy, "state", State.ACTIVE.value)
    # now `yep` should be enabled
    assert controller._button.isEnabled()

    controller.setEnabled(False)
    assert not controller._button.isEnabled()

    controller.setEnabled(True)
    assert controller._button.isEnabled()


def test_icon_command_icon_change(icon_command_setup, mocker):
    from ..icon_command import BUTTON_ICONS
    controller = icon_command_setup

    actions = controller.widget.actions()
    assert len(actions) == 1

    exec_path = ("karabogui.controllers.display.icon_command."
                 "IconSelectionDialog.exec")
    new_icon = "Left"

    def exec_dialog(self):
        self.icon = BUTTON_ICONS[new_icon]
        self.name = new_icon

    mocker.patch(exec_path, exec_dialog)
    actions[0].trigger()
    assert controller._icon == BUTTON_ICONS[new_icon]
    assert controller.model.icon_name == new_icon


def test_dialog_init(icon_command_dialog_setup):
    from ..icon_command import BUTTON_ICONS, NO_SELECTION

    dialog = icon_command_dialog_setup

    assert dialog.icon == NO_SELECTION
    assert dialog.name == ""

    model = dialog.list_view.model()
    for row in range(model.rowCount()):
        index = model.index(row, 0)
        name = index.data(Qt.DisplayRole)

        assert name in BUTTON_ICONS


def test_dialog_valid_double_click(icon_command_dialog_setup, mocker):
    from ..icon_command import BUTTON_ICONS
    new_icon = "Left"

    def mocked_returns(role):
        if role == Qt.DisplayRole:
            return new_icon
        if role == Qt.UserRole + 1:
            return BUTTON_ICONS[new_icon]

    dialog = icon_command_dialog_setup

    mocked_index = mocker.Mock()
    mocked_index.data.side_effect = mocked_returns
    dialog.handleDoubleClick(mocked_index)

    assert dialog.icon == BUTTON_ICONS[new_icon]
    assert dialog.name == new_icon


def test_dialog_invalid_double_click(icon_command_dialog_setup, mocker):
    from ..icon_command import NO_SELECTION

    def mocked_returns(role):
        if role == Qt.DisplayRole:
            return None
        if role == Qt.UserRole + 1:
            return None

    dialog = icon_command_dialog_setup

    mocked_index = mocker.Mock()
    mocked_index.data.side_effect = mocked_returns
    dialog.handleDoubleClick(mocked_index)

    assert dialog.icon == NO_SELECTION
    assert dialog.name == ""



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from traits.api import Any, ArrayOrNone, Enum, HasStrictTraits, Int

from karabo.native import Encoding
from karabogui.controllers.arrays import (
    DIMENSIONS, get_dimensions_and_encoding, get_image_data, get_jpeg_data)
from karabogui.graph.common.api import Axes

# Special image dimensions
YUV422 = 2
YUV444 = 3
YUV_TYPES = {Encoding.YUV444, Encoding.YUV422_YUYV,
             Encoding.YUV422_UYVY}


class KaraboImageNode(HasStrictTraits):
    dim_x = Int
    dim_y = Int
    dim_z = Int
    encoding = Enum(*Encoding)
    stack_axis = Any

    # Internal traits
    _data = ArrayOrNone

    def set_value(self, image_node):
        dim_x, dim_y, dim_z, encoding = get_dimensions_and_encoding(image_node)

        self.dim_x = dim_x if dim_x is not None else 0
        self.dim_y = dim_y if dim_y is not None else 0
        self.dim_z = dim_z if dim_z is not None else 0
        self.encoding = encoding
        if encoding == Encoding.JPEG:
            self._data = get_jpeg_data(image_node, self.dim_z)
        else:
            self._data = get_image_data(image_node,
                                        self.dim_x, self.dim_y, self.dim_z)

        stack_axis = None
        if "stackAxis" in image_node:
            stack_axis = image_node.stackAxis.value
        self.stack_axis = stack_axis

    def get_data(self):
        # Get data depending on the image schema and encoding
        if self.stack_axis is not None:
            return self.get_slice(self.stack_axis, cell=0)
        elif self.encoding == Encoding.GRAY:
            return self.get_slice(DIMENSIONS['Z'], cell=0)
        else:
            return self._data

    def get_axis_dimension(self, axis):
        """Return the correct dimension from a given axis"""
        if axis == Axes.Y:
            return self.dim_y
        if axis == Axes.X:
            return self.dim_x
        if axis == Axes.Z:
            return self.dim_z
        else:
            raise ValueError(f'Invalid axis value: {axis}')

    def get_slice(self, axis, cell):
        """Returns the cell slice from the given axis"""
        if not self.dim_z or self._data is None:
            return self._data
        else:
            # Numpy arrays are received as x, y, z
            if axis == Axes.Y:
                return self._data[cell, :, :]
            if axis == Axes.X:
                return self._data[:, cell, :]
            if axis == Axes.Z:
                return self._data[:, :, cell]

    @property
    def is_valid(self):
        if self._data is None or self.dim_x < 1 or self.dim_y < 1:
            return False
        # Handle YUV encoding type
        if self.encoding in YUV_TYPES:
            if self.dim_z == YUV422:
                # input image is (u1, y1, v1, y2, u2, y3, v2, y4, ...)
                # only display "luma" (Y') component in the GUI
                self._data = self._data[:, :, 1]
                self.dim_z = 0
                self.encoding = Encoding.GRAY
            elif self.dim_z == YUV444:
                # input image is (u1, y1, v1, u2, y2, v2, ...)
                # only display "luma" (Y') component in the GUI
                self._data = self._data[:, :, 1]
                self.dim_z = 0
                self.encoding = Encoding.GRAY
            else:
                return False

        return True



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on February 18, 2020
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy.QtCore import QRect, Slot
from qtpy.QtGui import QColor
from qtpy.QtWidgets import QAction, QDialog, QPlainTextEdit

from karabogui.dialogs.api import GREY, StickerDialog
from karabogui.fonts import get_qfont
from karabogui.widgets.hints import KaraboSceneWidget


class StickerWidget(KaraboSceneWidget, QPlainTextEdit):
    """A `StickerWidget` for creating editable text widgets on the `scene`"""

    def __init__(self, model, parent=None):
        super().__init__(model=model, parent=parent)
        self.set_widget_properties(model)
        self.setGeometry(QRect(model.x, model.y, model.width, model.height))
        self.setReadOnly(True)

        # Enforce the same frame style as with the dialog text edit to
        # have the same scroll area dimensions
        self.setFrameShape(QPlainTextEdit.Box | QPlainTextEdit.Plain)
        edit_action = QAction("Edit Sticker", self)
        edit_action.triggered.connect(self.edit_colors_text)
        self.addAction(edit_action)

    def add_proxies(self, proxies):
        """Satisfy the informal widget interface."""

    def apply_changes(self):
        """Satisfy the informal widget interface."""

    def decline_changes(self):
        """Satisfy the informal widget interface."""

    def destroy(self):
        """Satisfy the informal widget interface."""
        super().destroy()

    def set_visible(self, visible):
        """Satisfy the informal widget interface."""

    def update_global_access_level(self, level):
        """Satisfy the informal widget interface."""

    def set_geometry(self, rect):
        """Satisfy the informal widget interface."""
        self.model.trait_set(x=rect.x(), y=rect.y(),
                             width=rect.width(), height=rect.height())
        self.setGeometry(rect)

    def translate(self, offset):
        new_pos = self.pos() + offset
        self.model.trait_set(x=new_pos.x(), y=new_pos.y())
        self.move(new_pos)

    @Slot()
    def edit_colors_text(self):
        parent = self.parent()
        self.edit(parent)

    def edit(self, scene_view):
        dialog = StickerDialog(self.model, parent=scene_view)
        if dialog.exec() == QDialog.Rejected:
            return
        model = dialog.model
        self.model.trait_set(text=model.text,
                             font=model.font, background=model.background,
                             foreground=model.foreground)
        self.set_widget_properties(model)

    def set_widget_properties(self, model):
        """Set the new widget properties according to the model"""
        self.setPlainText(model.text)

        # We have to use the effective font string to correct font size
        font_string = get_qfont(model.font).toString()

        sheet = []
        sheet.append(f'qproperty-font: "{font_string}";')
        sheet.append(f'color: "{model.foreground}";')
        sheet.append(f'background-color: "{model.background}";')

        # Add borders
        bg = model.background
        if bg == 'transparent':
            bg = GREY
        color = QColor(bg).darker(120).name()
        sheet.append('border: 1px solid;')
        sheet.append('border-top: 5px solid;')
        sheet.append(f'border-color: {color};')

        self.setStyleSheet("QPlainTextEdit {{ {} }}".format("".join(
            sheet)))



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.

from karabo.common.api import set_initialized_flag
from karabo.common.project.api import ProjectModel
from karabogui.project.restore import ProjectRestoreSession
from karabogui.testing import assert_trait_change


def test_restore_project():
    """A simple test that the finished flag is fired once the project is
    initialized"""
    project = ProjectModel(simple_name="testproject")
    restore = ProjectRestoreSession(project=project)
    with assert_trait_change(restore, "finished"):
        # Reload the project
        set_initialized_flag(project, True)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np
from qtpy.QtCore import QPointF

from karabogui.graph.common.api import AuxPlots, ROITool
from karabogui.testing import GuiTestCase

from ..base import KaraboImageView


class TestKaraboImageView(GuiTestCase):

    def setUp(self):
        super().setUp()
        self.widget = KaraboImageView()
        self.roi = self.widget.add_roi()
        self.aux_plots = self.widget.add_aux(AuxPlots.ProfilePlot)

    def tearDown(self):
        self.roi.destroy()
        self.aux_plots.destroy()
        self.widget.destroy()
        self.widget = None

    def test_first_image(self):
        """ This test checks when the image is set,
            the aux plots should have data """

        # Check no data on init
        assert not self.widget.plotItem.image_set
        assert self.aux_plots.current_plot == AuxPlots.NoPlot

        # Setup widget
        self.widget.show_aux_plots(AuxPlots.ProfilePlot)
        roi_item = self.roi.add(ROITool.Rect, pos=QPointF(3, 3), size=(3, 3))
        self.roi.show(ROITool.Rect)
        assert roi_item.isVisible()

        self._assert_aux_plots_data(has_data=False)

        # Set image
        image = np.ones((150, 100))
        self.widget.plotItem.set_image(image)

        # Check if image is properly set
        assert self.widget.plotItem.image_set
        assert np.array_equal(self.widget.plotItem.image, image)

        self._assert_aux_plots_data(has_data=True)

    def test_empty_np_image(self):
        """This test checks empty image with np.empty as input"""
        # Set image
        image = np.empty((100, 100))
        self.widget.plotItem.set_image(image)

        # Check if image is properly set
        assert self.widget.plotItem.image_set
        np.testing.assert_array_equal(self.widget.plotItem.image, image)

        assert self.aux_plots.current_plot == AuxPlots.NoPlot

    def test_empty_list_image(self):
        """This test checks empty image with [] as input"""
        # Set image
        image = np.array([])
        self.widget.plotItem.set_image(image)

        # Check if image is properly set
        assert not self.widget.plotItem.image_set

    def test_none_image(self):
        """This test checks empty image with None as input"""
        # Set image
        image = None
        self.widget.plotItem.set_image(image)

        # Check if image is properly set
        assert not self.widget.plotItem.image_set

    def _assert_aux_plots_data(self, has_data=True):
        current_plot = self.aux_plots.current_plot
        assert current_plot is not None

        assert_method = (self.assertFalse if has_data
                         else self.assertTrue)
        aggregator = self.aux_plots._aggregators[current_plot]
        for controller in aggregator.controllers.values():
            data_item = controller.plot._data_item
            assert_method(data_item.xData is None or len(data_item.xData) == 0)
            assert_method(data_item.xData is None or len(data_item.xData) == 0)

    def test_axes_labels(self):
        # Restore config
        self._assert_axes_labels(x_label='', x_units='',
                                 y_label='', y_units='')
        self._assert_axes_labels(x_label='Foo', x_units='',
                                 y_label='', y_units='')
        self._assert_axes_labels(x_label='', x_units='',
                                 y_label='Bar', y_units='')

    def _assert_axes_labels(self, **config):
        # Get axis items
        top = self.widget.plotItem.getAxis("top")
        bottom = self.widget.plotItem.getAxis("bottom")
        left = self.widget.plotItem.getAxis("left")
        right = self.widget.plotItem.getAxis("right")

        # Setup expected text and visibility
        x_label = f'{config["x_label"]}'
        if config["x_units"]:
            x_label += f' ({config["x_units"]})'
        x_visible = bool(config["x_label"] or config["x_units"])

        y_label = f'{config["y_label"]}'
        if config["y_units"]:
            y_label += f' ({config["y_units"]})'
        y_visible = bool(config["y_label"] or config["y_units"])

        # Apply config
        self.widget.restore(config)

        # Check blank axes
        assert top.labelText == x_label
        assert bottom.labelText == ''
        assert left.labelText == y_label
        assert right.labelText == ''

        # Check visibility
        assert top.label.isVisible() == x_visible
        assert bottom.label.isVisible() is False  # always false
        assert left.label.isVisible() == y_visible
        assert right.label.isVisible() is False  # always false



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from unittest.mock import patch

from traits.api import pop_exception_handler, push_exception_handler

from karabo.common.project.api import MacroModel, ProjectModel
from karabo.common.scenemodel.api import SceneModel
from karabogui.project.controller.project_groups import (
    _load_macro, _load_scene)
from karabogui.singletons.project_model import ProjectViewItemModel
from karabogui.testing import GuiTestCase


class TestProjectGroupController(GuiTestCase):

    def setUp(self):
        super().setUp()
        push_exception_handler(lambda *args: None, reraise_exceptions=True)
        self.qt_model = ProjectViewItemModel(parent=None)

        # Instantiate project controller
        proj_model = ProjectModel()
        self.qt_model.root_model = proj_model
        self.proj_controller = self.qt_model.root_controller

    def tearDown(self):
        super().tearDown()
        pop_exception_handler()
        self.qt_model = None

    def test_load_macro(self):
        self._assert_load_macro_name("hello", is_valid=True)
        self._assert_load_macro_name("foo@bar.baz", is_valid=False)

    def test_load_scene(self):
        self._assert_load_scene_name("hello", is_valid=True)
        self._assert_load_scene_name("foo@bar.baz", is_valid=False)

    def _assert_load_macro_name(self, name, is_valid):
        patched_get_config = patch(
            "karabogui.project.controller.project_groups.get_config",
            return_value={"data_dir": ""})

        patched_get_filename = patch(
            "karabogui.project.controller.project_groups.getOpenFileName",
            return_value=name + ".py")

        patched_read_model = patch(
            "karabogui.project.controller.project_groups.read_macro",
            return_value=MacroModel())

        patched_show_filename_error = patch(
            "karabogui.project.controller.project_groups.show_filename_error")

        with patched_get_config, patched_get_filename, \
                patched_read_model, patched_show_filename_error:
            _load_macro(self.proj_controller)

        macros = [macro.simple_name
                  for macro in self.proj_controller.model.macros]

        if is_valid:
            assert name in macros
        else:
            assert name not in macros

    def _assert_load_scene_name(self, name, is_valid):
        patched_get_config = patch(
            "karabogui.project.controller.project_groups.get_config",
            return_value={"data_dir": ""})

        patched_get_filename = patch(
            "karabogui.project.controller.project_groups.getOpenFileName",
            return_value=name + ".svg")

        patched_read_scene = patch(
            "karabogui.project.controller.project_groups.read_scene",
            return_value=SceneModel())

        patched_show_filename_error = patch(
            "karabogui.project.controller.project_groups.show_filename_error")

        with patched_get_config, patched_get_filename, \
                patched_read_scene, patched_show_filename_error:
            _load_scene(self.proj_controller)

        scenes = [scene.simple_name
                  for scene in self.proj_controller.model.scenes]

        if is_valid:
            assert name in scenes
        else:
            assert name not in scenes



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on July 9, 2011
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import os
import pickle
import socket

from ipykernel.inprocess.client import InProcessKernelClient
from qtconsole import inprocess, kernel_mixins
from qtconsole.rich_jupyter_widget import RichJupyterWidget

from karabo.common.states import State
from karabo.native import Hash
from karabogui import messagebox
from karabogui.binding.api import PropertyProxy, get_binding_value
from karabogui.request import send_property_changes
from karabogui.singletons.api import get_network, get_topology
from karabogui.topology.api import get_macro_servers


class IPythonWidget(RichJupyterWidget):
    def __init__(self, banner=None, call_backs={}, *args, **kwargs):
        if banner is not None:
            self.banner = banner
        super().__init__(*args, local_kernel=True, **kwargs)

        self.kernel_manager = KernelManager()
        self.kernel_manager.start_kernel()
        self.kernel_client = self.kernel_manager.client()

        session_start_cb = call_backs.get("start_ch")
        session_stop_cb = call_backs.get("stop_ch")
        if session_start_cb:
            self.kernel_client.started_channels.connect(session_start_cb)
        if session_stop_cb:
            self.kernel_client.stopped_channels.connect(session_stop_cb)

    def stop(self):
        self.kernel_client.stop_channels()
        self.kernel_manager.shutdown_kernel()

    def destroy(self):
        # Cleanup first
        self.kernel_client.started_channels.disconnect()
        self.kernel_client.stopped_channels.disconnect()
        self.deleteLater()
        super().destroy()


class Channel(inprocess.QtInProcessChannel):
    def __init__(self, *args):
        super().__init__(*args)
        self.last = None

    def is_alive(self):
        return self.client.alive

    def receive(self, value):
        if value != self.last:
            self.last = value
            self.call_handlers_later(pickle.loads(value))


class KernelManager(kernel_mixins.QtKernelManagerMixin):
    __client = None

    def start_kernel(self):
        macro_servers = get_macro_servers()
        if not macro_servers:
            messagebox.show_error(
                "No Macro server found in system topology. "
                "Console cannot be started.")
            raise OSError

        hostname = socket.gethostname().replace(".", "_")
        network = get_network()
        self.name = f"CLI-{hostname}-{os.getpid()}"
        serverId = macro_servers[0]
        network.onInitDevice(serverId, "IPythonKernel", self.name, Hash())

    def shutdown_kernel(self):
        network = get_network()
        network.onKillDevice(self.name)

    def client(self):
        if self.__client is None:
            self.__client = KernelClient(self.name)
        return self.__client

    def interrupt_kernel(self):
        self.__client.interrupt.execute()


class KernelClient(inprocess.QtInProcessKernelClient):
    shell_channel_class = Channel
    iopub_channel_class = Channel
    stdin_channel_class = Channel
    hb_channel_class = inprocess.QtInProcessHBChannel

    def __init__(self, name):
        super().__init__()
        self.alive = False
        self.started = False
        self.name = name

        device = get_topology().get_device(self.name)
        self.device = device
        self.interrupt = PropertyProxy(root_proxy=device, path="interrupt")
        self._iopub = PropertyProxy(root_proxy=device, path="iopub")
        self._shell = PropertyProxy(root_proxy=device, path="shell")
        self._stdin = PropertyProxy(root_proxy=device, path="stdin")

        device.on_trait_change(self._state_update,
                               "state_binding.config_update")
        for prop in (self._iopub, self._shell, self._stdin):
            prop.on_trait_change(self._proxy_update, "binding.config_update")
            # The device is killed when closing. No need to stop_monitoring()
            prop.start_monitoring()

    def _proxy_update(self, binding, name, event):
        # This is bound to "binding.config_update" notifications
        if name != "config_update":
            return
        mapping = {
            self._iopub: self.iopub_channel,
            self._shell: self.shell_channel,
            self._stdin: self.stdin_channel,
        }
        for proxy, channel in mapping.items():
            if proxy.binding is binding:
                value = get_binding_value(proxy)
                if value is not None:
                    channel.receive(value)
                break

    def _state_update(self, binding, name, event):
        if name != "config_update":
            return

        value = get_binding_value(binding, State.UNKNOWN)
        self.alive = State(value) is State.STARTED
        if self.alive:
            if not self.started:
                self.start_channels()
                self.started = True
                self.started_channels.emit()

    def _dispatch_to_kernel(self, msg):
        self._shell.edit_value = pickle.dumps(msg)
        send_property_changes([self._shell])

    def start_channels(self):
        # magic: skip InProcessKernelClient, does stuff we don"t want
        super(InProcessKernelClient, self).start_channels()

    def input(self, string):
        msg = self.session.msg("input_reply", {"value": string})
        self._stdin.edit_value = pickle.dumps(msg)
        send_property_changes([self._stdin])



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import pytest

from ..configuration import (
    BACKBONE, DIRECTORIES, NETWORK, PROJECT, USER, Configuration)


class MockSettings:
    last_path = None

    @classmethod
    def value(cls, path):
        return getattr(cls, path, None)

    @classmethod
    def setValue(cls, name, value):
        setattr(cls, name, value)

    @classmethod
    def remove(cls, key):
        cls.last_path = key
        setattr(cls, key, None)

    def __call__(self, *args, **kwargs):
        return self


def test_configuration_namespace():
    config = Configuration()
    config['broker_topic'] = 'FXE'
    config['domain'] = 'CAS_INTERNAL'
    assert len(config) == 19
    assert config['domain'] == 'CAS_INTERNAL'

    keys = [
        'access_level',
        'broker_topic',
        'check_updates',
        'data_dir',
        'development',
        'domain',
        'gui_servers',
        'highDPI',
        'logbook_header_style',
        'logbook_stream',
        'logbook_topic',
        'macro_development',
        'main_geometry',
        'project_sort_column',
        'project_sort_order',
        'refresh_token',
        'refresh_token_user',
        'username',
        'wizard']

    assert list(config.keys()) == keys


def test_set_wrong_key():
    config = Configuration()
    with pytest.raises(KeyError):
        config['noitem'] = "Beep"

    with pytest.raises(KeyError):
        config.noitem = "Boom"


def test_default_value():
    """Every item that specifies a data type must have a default value"""
    config = Configuration()
    counter = 0
    for key in config.keys():
        item = getattr(config.__class__, key)
        if item.dtype is not None:
            counter += 1
            assert item.default is not None
    assert counter == 6


def test_set_bool_value(mocker):
    target = 'karabogui.singletons.configuration.QSettings'
    mocker.patch(target, new=MockSettings)
    config = Configuration()
    assert config["wizard"] is True
    config["wizard"] = False
    assert config["wizard"] is False


def test_erase_value(mocker):
    target = 'karabogui.singletons.configuration.QSettings'
    mock = MockSettings()
    mocker.patch(target, new=mock)
    config = Configuration()
    assert config["wizard"] is True
    config["wizard"] = False
    del config["wizard"]
    # Nothing happens for cached value
    assert config["wizard"] is False
    assert mock.last_path == "user/wizard"
    # Shared item
    assert config["refresh_token"] is None
    config["refresh_token"] = "12345"
    assert config["refresh_token"] == "12345"
    del config["refresh_token"]
    assert mock.last_path == "authentication/refresh_token"
    assert config["refresh_token"] is None


def test_configuration_groups_info():
    config = Configuration()
    groups = config.groups()
    assert len(groups) == 5
    user_group = [item.name for item in groups[USER]]
    assert len(user_group) == 10
    assert 'wizard' in user_group
    assert 'main_geometry' in user_group
    assert 'highDPI' in user_group
    assert 'development' in user_group
    assert 'project_sort_column' in user_group
    assert 'project_sort_order' in user_group
    network_group = [item.name for item in groups[NETWORK]]
    assert len(network_group) == 4
    assert 'access_level' in network_group
    assert 'username' in network_group
    assert 'gui_servers' in network_group
    project_group = [item.name for item in groups[PROJECT]]
    assert len(project_group) == 1
    assert 'domain' in project_group
    dir_group = [item.name for item in groups[DIRECTORIES]]
    assert len(dir_group) == 1
    assert 'data_dir' in dir_group
    bone_group = [item.name for item in groups[BACKBONE]]
    assert len(bone_group) == 1
    assert 'broker_topic' in bone_group

    info = config.info()
    assert USER in info



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on February 1, 2012
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import logging
import uuid
from functools import wraps
from inspect import signature
from weakref import WeakKeyDictionary, WeakValueDictionary

from qtpy.QtCore import QObject, Slot
from qtpy.QtWidgets import QMessageBox

import karabogui.access as krb_access
from karabo.native import AccessLevel, AccessMode, Assignment, Hash, Timestamp
from karabogui import messagebox
from karabogui.background import Priority, executeLater
from karabogui.binding.api import (
    ProxyStatus, apply_fast_data, extract_configuration)
from karabogui.const import KARABO_CLIENT_ID
from karabogui.events import KaraboEvent, broadcast_event
from karabogui.logger import get_logger
from karabogui.singletons.api import get_config, get_network, get_topology
from karabogui.util import get_reason_parts, move_to_cursor, show_wait_cursor

from .util import realign_topo_hash


def project_db_handler(fall_through=False):
    """Decorator for project DB handlers to cut down on boilerplate"""

    def inner(handler):
        @wraps(handler)
        def wrapped(self, success, request, reply, reason=""):
            if not success:
                error, details = get_reason_parts(reason)
                messagebox.show_error(error, details=details)
            if fall_through or success:
                return handler(self, success, request, reply, reason=reason)

        return wrapped

    return inner


def getLevelName(level):
    """Convert a logging level from the gui server to an integer"""
    return logging.getLevelName(level.upper())


class Manager(QObject):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        # this dictionary maps the deviceIds to a weak reference of the
        # DeviceProxy instances waiting for a reply from the server
        self._waiting_devices = WeakValueDictionary()
        # this dictionary maps a weak reference DeviceProxy with the list of
        # property proxies waiting to be heard
        self._waiting_properties = WeakKeyDictionary()
        # The system topology singleton
        self._topology = get_topology()
        # A dict which includes big networkData
        self._big_data = {}
        # Handler callbacks for `handle_requestGeneric`
        self._request_handlers = {}

        # Awaiting device instantiations
        self._awaiting_instantiations = set()
        self._show_big_data_proc = False
        network = get_network()
        network.signalServerConnectionChanged.connect(
            self.onServerConnectionChanged)
        network.signalReceivedData.connect(self.onReceivedData)

        if get_config()["development"]:
            self.toggleBigDataPerformanceMonitor()

    def toggleBigDataPerformanceMonitor(self):
        self._show_big_data_proc = True

    def expect_properties(self, device_proxy, properties):
        """Register the property proxies as waiting for replies"""
        device_id = device_proxy.device_id
        self._waiting_devices[device_id] = device_proxy
        proxies = self._waiting_properties.get(device_proxy, [])
        proxies.extend(properties)
        self._waiting_properties[device_proxy] = proxies

    def initDevice(self, serverId, classId, deviceId, config=None):
        """Instantiate a project device"""
        proxy = self._topology.get_project_device_proxy(
            deviceId, serverId, classId)
        if proxy is None:
            # This should never happen!
            return

        def _instantiate_device():
            """Instantiate a project device"""
            nonlocal config

            if proxy.online:
                # Note: Device was instantiated from a different place
                # and not us, conflict! This can happen especially from an
                # awaiting instantiation!
                messagebox.show_error(
                    f"Aborting instantiation for device {deviceId} because it "
                    "is already online!")
                return

            if config is None:
                config = extract_configuration(proxy.binding)

            schema = self._topology.get_schema(serverId, classId)
            if schema is None:
                # XXX: This should not happen anymore. However, we go safe here
                messagebox.show_error(
                    f"Aborting instantiation for device {deviceId} due to "
                    "a missing schema!")
                return

            # Remove all obsolete paths which are not anymore in the schema
            # Note: This is supposed to throw on the device server level
            obsolete_paths = [pth for pth, _, _ in Hash.flat_iterall(config)
                              if pth not in schema.hash]
            for key in obsolete_paths:
                config.erase(key)

            # Remove all read only and assignment internal paths
            readonly_paths = [pth for pth, _, _ in Hash.flat_iterall(config)
                              if schema.hash[pth, "accessMode"] ==
                              AccessMode.READONLY.value or
                              schema.hash[pth, "assignment"] ==
                              Assignment.INTERNAL.value]
            for key in readonly_paths:
                config.erase(key)

            # Send signal to network
            get_network().onInitDevice(serverId, classId, deviceId, config)

        if len(proxy.binding.value) > 0:
            _instantiate_device()
        else:
            # We don't have a schema and have to defer the instantiation.
            # We are sneaky and wait until the offline configuration has
            # been applied!
            get_logger().info("Requesting class schema with classId "
                              f"<b>{classId}</b> for <b>{deviceId}</b> "
                              "and deferring instantiation")

            def _config_update_handler():
                proxy.on_trait_change(_config_update_handler, "config_update",
                                      remove=True)
                if deviceId in self._awaiting_instantiations:
                    self._awaiting_instantiations.discard(deviceId)
                    _instantiate_device()

            self._awaiting_instantiations.add(deviceId)
            proxy.on_trait_change(_config_update_handler, "config_update")
            proxy.ensure_class_schema()

    def callDeviceSlot(self, handler, instance_id, slot_name, params):
        """Call a device slot using the `requestGeneric` mechanism.

        See karabogui.request for more details.
        """
        # Generate a unique token for the transaction
        token = uuid.uuid4().hex
        assert token not in self._request_handlers

        # Set the callback handler
        assert callable(handler)
        self._request_handlers[token] = handler
        # Call the GUI server
        get_network().onExecuteGeneric(instance_id, slot_name, params,
                                       token=token)
        return token

    def shutdownDevice(self, deviceId, showConfirm=True, parent=None):
        if showConfirm:
            ask = ('Do you really want to shutdown the device '
                   '"<b>{}</b>"?').format(deviceId)
            msg_box = QMessageBox(QMessageBox.Question, 'Shutdown device', ask,
                                  QMessageBox.Yes | QMessageBox.No,
                                  parent=parent)

            msg_box.setModal(False)
            msg_box.setDefaultButton(QMessageBox.No)
            move_to_cursor(msg_box)
            if msg_box.exec() == QMessageBox.No:
                return

        # We can also shutdown a device with `Shutdown all devices` while
        # waiting for their instantiation. We make sure to delete our request
        # here!
        self._awaiting_instantiations.discard(deviceId)

        get_network().onKillDevice(deviceId)

    def shutdownServer(self, serverId, parent=None):
        ask = ('Do you really want to shutdown the server '
               '"<b>{}</b>"?').format(serverId)
        msg_box = QMessageBox(QMessageBox.Question, 'Shutdown server', ask,
                              QMessageBox.Yes | QMessageBox.No, parent=parent)
        msg_box.setModal(False)
        msg_box.setDefaultButton(QMessageBox.No)
        move_to_cursor(msg_box)
        if msg_box.exec() == QMessageBox.No:
            return
        # Note: We shutdown a server although we want to instantiate devices
        # from there.
        if self._awaiting_instantiations:
            deviceIds = {deviceId for deviceId, project_device
                         in self._topology._project_devices.items()
                         if project_device.server_id == serverId}
            discard = deviceIds.intersection(self._awaiting_instantiations)
            for deviceId in discard:
                self._awaiting_instantiations.discard(deviceId)
        get_network().onKillServer(serverId)

    @Slot(object)
    def onReceivedData(self, hsh):
        handler = getattr(self, "handle_" + hsh["type"])
        kwargs = {k: v for k, v in hsh.items() if k != "type"}
        handler(**kwargs)

    @Slot(bool)
    def onServerConnectionChanged(self, isConnected):
        """If the server connection is changed, the model needs an update.
        """
        # Broadcast event to all panels which need a reset
        broadcast_event(KaraboEvent.NetworkConnectStatus,
                        {'status': isConnected})

        # Clear the system topology
        if not isConnected:
            self._topology.clear()

            # Delete any request of instantiating devices
            self._awaiting_instantiations.clear()

            # Any pending requests are effectively timed-out and removed
            pending = len(self._request_handlers)
            if pending:
                get_logger().error(
                    f"Erased <b>{pending}</b> pending generic requests due to "
                    "client disconnect.")
            self._request_handlers.clear()

    def handle_getGuiSessionInfo(self, **info):
        """Handle the user session information from the gui server"""
        success = info["success"]
        reason = info["reason"]
        if not success:
            messagebox.show_error(
                "No UserSession Information", details=reason)
            return

        data = {"sessionStartTime": info.get("sessionStartTime"),
                "sessionDuration": info.get("sessionDuration")}
        broadcast_event(KaraboEvent.UserSessionInfo, data)

    def handle_setLogLevelReply(self, **info):
        """Handle the log priority reconfiguration reply of the server"""
        if not info.get("success", True):
            reason = info.get("reason")
            error, details = get_reason_parts(reason)
            input_info = info["input"]
            instanceId = input_info["instanceId"]
            level = input_info["level"]
            log_text = (f"Log level reconfiguration of <b>{instanceId}</b>"
                        f" with priority <b>{level}</b> failed.<br>"
                        f"The reason is:<br><i>{error}</i>")
            get_logger().error(log_text)
            messagebox.show_error(log_text, details=details)

    def handle_reconfigureReply(self, **info):
        """Handle the reconfigure reply of the gui server"""
        success = info['success']
        input_info = info['input']
        deviceId = input_info['deviceId']
        if not success:
            reason = info.get('failureReason') or info.get('reason')
            reason, details = get_reason_parts(reason)
            # clear the waiting queue
            device = self._waiting_devices.pop(deviceId, None)
            if device is None:
                return
            props = self._waiting_properties.pop(device, [])
            paths = set(input_info['configuration'].paths())
            pending = [property_proxy for property_proxy in props
                       if property_proxy.path not in paths]
            if pending:
                # in case multiple updates are sent and only few failed
                self.expect_properties(device, pending)
            text = (f'Device reconfiguration of <b>{deviceId}</b> encountered '
                    'an error. <br><br>The reason is:<br>'
                    f'<i>{reason}</i><br>')
            if details is not None:
                text += '<br>Click "Show Details..." for more information.'
            # Provide a text for the logger!
            log_text = (f"Device reconfiguration of <b>{deviceId}</b> with "
                        f"properties <b>{', '.join(paths)}</b> failed")
            get_logger().error(log_text)
            messagebox.show_error_at_cursor(text, details=details)
        else:
            device_proxy = self._waiting_devices.pop(deviceId, None)
            if device_proxy is None:
                return
            prop_proxies = self._waiting_properties.pop(device_proxy, [])
            paths = set(input_info['configuration'].paths())
            pending = []
            for prop_proxy in prop_proxies:
                if prop_proxy.path not in paths:
                    pending.append(prop_proxy)
                    continue
                # Erase the successful configured edit_value and wait for the
                # config update
                prop_proxy.trait_setq(edit_value=None)
            if pending:
                # in case multiple updates are sent, wait for the pending ones
                self.expect_properties(device_proxy, pending)
            broadcast_event(
                KaraboEvent.UpdateValueConfigurator,
                {'proxy': device_proxy})

    def handle_requestGeneric(self, success, request=None, reply=None,
                              reason=''):
        """Handle the requestGeneric reply from the GUI server

        Generic requests are supposed to have a `token` in the input arguments
        Unfolding the `token` should provide a request handler

        """
        token = request["token"]
        handler = self._request_handlers.pop(token, None)
        if handler is None:
            return
        reply = reply if success is True else reason
        # Wrap the handler invocation in an exception swallower
        try:
            sig = signature(handler)
            if len(sig.parameters) == 2:
                handler(success, reply)
            else:
                handler(success, reply, request)
        except Exception as ex:
            # But at least show something when exceptions are being swallowed
            msg = ('Exception of type "{}" caught in callback handler "{}" '
                   'with reply:\n{}')
            print(msg.format(type(ex).__name__, handler, reply))

    def handle_log(self, messages):
        broadcast_event(KaraboEvent.LogMessages, {'messages': messages})

    def handle_configurationFromPast(self, **info):
        success = info['success']
        # NOTE: Show a reasonable time format!
        time = info['time']
        time = Timestamp(time).toLocal()
        deviceId = info['deviceId']
        if not success:
            reason = info['reason']
            reason, details = get_reason_parts(reason)
            msg = (f"The configuration of `{deviceId}` requested at time point"
                   f" `{time}` was not retrieved!<br><br>"
                   "The reason is:<br>"
                   f"<i>{reason}</i>")
            messagebox.show_error(msg, details=details)
            return

        config = info['config']
        config_time = info['configTimepoint']
        config_time = Timestamp(config_time).toLocal()
        time_match = info['configAtTimepoint']

        # We might deal with old gui server < 2.11
        preview = info.get('preview', False)
        broadcast_event(KaraboEvent.ShowConfigurationFromPast,
                        {'deviceId': deviceId, 'configuration': config,
                         'time': time, 'config_time': config_time,
                         'preview': preview, 'time_match': time_match})

    def handle_brokerInformation(self, **info):
        self._set_server_information(info)

    def handle_serverInformation(self, **info):
        self._set_server_information(info)

    def handle_loginInformation(self, **info):
        """Handle the login information from the gui server

        If the gui server is in readOnly mode, the readOnly boolean is included
        here as well since 2.20.X.
        """
        get_config()['username'] = info["username"]
        read_only = info.get("readOnly", False)
        krb_access.set_read_only_mode(read_only)
        if read_only:
            access = AccessLevel.OBSERVER
        else:
            access = AccessLevel(info["accessLevel"])
        # Reset the notice of session
        if krb_access.SESSION_END_NOTICE:
            krb_access.SESSION_END_NOTICE = False
            broadcast_event(KaraboEvent.UserSession, {})

        krb_access.HIGHEST_ACCESS_LEVEL = access
        if krb_access.GLOBAL_ACCESS_LEVEL != access:
            krb_access.GLOBAL_ACCESS_LEVEL = access
            broadcast_event(KaraboEvent.AccessLevelChanged, {})

        broadcast_event(KaraboEvent.LoginUserChanged, {})

    def handle_onBeginTemporarySession(self, **info):
        """Handle the response from gui server on starting the temporary
        session request by the user"""
        if info["success"]:
            krb_access.TEMPORARY_SESSION_USER = info.get("username")
            krb_access.TEMPORARY_SESSION_WARNING = False
            access_level = AccessLevel(info["accessLevel"])

            krb_access.HIGHEST_ACCESS_LEVEL = access_level
            if krb_access.GLOBAL_ACCESS_LEVEL != access_level:
                krb_access.GLOBAL_ACCESS_LEVEL = access_level
                broadcast_event(KaraboEvent.AccessLevelChanged, {})

            broadcast_event(KaraboEvent.LoginUserChanged, {})
            broadcast_event(KaraboEvent.UserSession, {})

    def handle_onEndTemporarySession(self, **info):
        """Handle the response from gui server on end temporary session
        request by user"""
        self._end_temporary_session(info)

    def handle_onTemporarySessionExpired(self, **info):
        """Handle the temporary session expired message from gui server """
        self._end_temporary_session(info)

    def handle_onEndSessionNotice(self, **info):
        """Handle the session notice message from gui server """
        get_logger().critical(
            "End of session received, please reauthenticate. "
            f"You have {info['secondsToExpiration']} seconds left before "
            "being logged out.")
        krb_access.SESSION_END_NOTICE = True
        broadcast_event(KaraboEvent.UserSession, {})

    def handle_onSessionExpired(self, **info):
        """Handle the session expired message from gui server """
        broadcast_event(KaraboEvent.UserSession, {})
        messagebox.show_warning("Session has been terminated by GUI server.")

    def _end_temporary_session(self, info):
        level_before = info.get("levelBeforeTemporarySession")
        if level_before is None:
            return
        access_level = AccessLevel(level_before)
        if krb_access.HIGHEST_ACCESS_LEVEL != access_level:
            krb_access.HIGHEST_ACCESS_LEVEL = access_level

        if krb_access.GLOBAL_ACCESS_LEVEL != access_level:
            krb_access.GLOBAL_ACCESS_LEVEL = access_level
            broadcast_event(KaraboEvent.AccessLevelChanged, {})

        krb_access.TEMPORARY_SESSION_USER = None
        broadcast_event(KaraboEvent.LoginUserChanged, {})
        broadcast_event(KaraboEvent.UserSession, {})

    def handle_onEndTemporarySessionNotice(self, **info):
        """Broadcast before the temporary session ends"""
        krb_access.TEMPORARY_SESSION_WARNING = True
        broadcast_event(KaraboEvent.UserSession, {})

    @show_wait_cursor
    def handle_systemTopology(self, systemTopology):
        systemTopology["server"] = realign_topo_hash(
            systemTopology["server"], "host")
        self._topology.initialize(systemTopology)

        devices, servers = _extract_topology_devices(systemTopology)
        broadcast_event(KaraboEvent.SystemTopologyUpdate,
                        {'devices': devices, 'servers': servers})

    def handle_systemVersion(self, **info):
        """Handle the version number reply from the GUI server"""
        pass

    def handle_topologyUpdate(self, changes):
        devices, servers = self._topology.topology_update(changes)

        gone_instanceIds = [instance_id for instance_id, _, _ in devices]

        # Update topology interested listeners!
        new_devices, new_servers = _extract_topology_devices(
            changes['new'])

        devices.extend(new_devices)
        servers.extend(new_servers)

        if update := changes["update"]:
            broadcast_event(KaraboEvent.SystemTopologyInstanceUpdate, update)

        broadcast_event(KaraboEvent.SystemTopologyUpdate,
                        {'devices': devices, 'servers': servers})

        broadcast_event(KaraboEvent.ClearConfigurator,
                        {'devices': gone_instanceIds})

    def handle_classSchema(self, serverId, classId, schema):
        self._topology.class_schema_updated(serverId, classId, schema)

    def handle_deviceSchema(self, deviceId, schema):
        proxy = self._topology.device_schema_updated(deviceId, schema)
        if proxy is None:
            return

        # Refresh the configurator iff this proxy is already showing
        broadcast_event(KaraboEvent.UpdateDeviceConfigurator,
                        {'proxy': proxy})

    def handle_deviceConfiguration(self, deviceId, configuration):
        self._topology.device_config_updated(deviceId, configuration)

    def handle_deviceConfigurations(self, configurations):
        """Update a bulk series of device configurations

        The arriving hash provides the deviceId's as keys and configurations
        as their values.
        """
        for deviceId, config in configurations.items():
            self._topology.device_config_updated(deviceId, config)

    def handle_executeReply(self, **info):
        """Handle the execute reply of the gui server"""
        success = info['success']
        if not success:
            reason = info.get('failureReason') or info.get('reason')
            reason, details = get_reason_parts(reason)
            input_info = info['input']
            deviceId = input_info['deviceId']
            command = input_info['command']
            text = (f'Execute slot <b>{command}</b> of device '
                    f'<b>{deviceId}</b> has encountered an error!'
                    '<br><br>'
                    'The reason is:<br>'
                    f'<i>{reason}</i>'
                    '<br>')
            if details is not None:
                text += '<br>Click "Show Details..." for more information.'
            messagebox.show_error(text, details=details)

    def handle_propertyHistory(self, **info):
        """The handler of the property history.

        The `info` dictionary contains:

            deviceId (str): the deviceId of the property
            property (str): the key of the property
            success (bool): success boolean of the request
            data (HashList): the data returned by the historic data request.

            failureReason (str): With Karabo <= 2.11, the failure reason is
                                 provided here.
            reason: Forward compatibility for future releases
        """
        success = info["success"]
        if success:
            deviceId = info["deviceId"]
            key = info["property"]
            data = info.get("data", None)
            if data is not None:
                device_proxy = self._topology.get_device(deviceId)
                device_proxy.publish_historic_data(key, data)
        else:
            # Provide a logger message on failure for developers
            deviceId = info["deviceId"]
            key = info["property"]
            text = (f"Historic data request for property <b>{key}</b> "
                    f"of device <b>{deviceId}</b> failed.")
            get_logger().debug(text)

    # ---------------------------------------------------------------------
    # Current Configuration Interface

    def handle_listInitConfigurations(self, success, request,
                                      reply, reason=''):
        """Handle the reply of the listInitConfigurations slot"""
        deviceId = request['args.deviceId']
        if not success:
            messagebox.show_error(f"Requesting a list of configurations for "
                                  f"{deviceId} failed!", details=reason)
            return

        broadcast_event(KaraboEvent.ListConfigurationUpdated,
                        {'items': reply['items'],
                         'deviceId': deviceId})

    def handle_getInitConfiguration(self, success, request, reply, reason=''):
        deviceId = request['args.deviceId']
        if not success:
            messagebox.show_error(f"Requesting a configuration for {deviceId} "
                                  "failed!", details=reason)
            return

        item = reply['item']
        configuration = item['config']
        broadcast_event(KaraboEvent.ShowInitConfiguration,
                        {'configuration': configuration,
                         'preview': request['preview'],
                         'classId': item["classId"],
                         'name': item['name'],
                         'deviceId': deviceId})

    def handle_deleteInitConfiguration(self, success, request,
                                       reply, reason=""):
        deviceId = request['args.deviceId']
        name = request['args.name']
        if not success:
            messagebox.show_error(f"Delete a configuration for {deviceId} "
                                  f"with name {name} failed!", details=reason)
            return

        get_network().onListInitConfigurations(deviceId)

    def handle_saveInitConfiguration(self, success, request, reply, reason=''):
        deviceId = request['args.deviceIds'][0]
        if not success:
            messagebox.show_error(f"Saving a configuration for {deviceId} "
                                  "failed!", details=reason)
            return

        if request.get('update', False):
            get_network().onListInitConfigurations(deviceId)

    # ---------------------------------------------------------------------
    # Current Project Interface

    @project_db_handler()
    def handle_projectListDomains(self, success, request, reply, reason=""):
        # ``reply`` is a Hash containing a list of domain names
        broadcast_event(KaraboEvent.ProjectDomainsList,
                        {'items': reply.get('domains', [])})

    @project_db_handler()
    def handle_projectListItems(self, success, request, reply, reason=""):
        broadcast_event(KaraboEvent.ProjectItemsList,
                        {'items': reply.get('items', [])})

    def handle_projectListProjectsWithItem(self, success, request,
                                           reply, reason=""):
        if not success:
            broadcast_event(
                KaraboEvent.ProjectFindWithItem,
                {'projects': [], 'error': reason})
        else:
            projects = reply.get('items', [])
            broadcast_event(
                KaraboEvent.ProjectFindWithItem, {
                    'items': projects, 'error': None})

    @project_db_handler(fall_through=True)
    def handle_projectLoadItems(self, success, request, reply, reason=""):
        d = {'success': success, 'items': reply.get('items', [])}
        broadcast_event(KaraboEvent.ProjectItemsLoaded, d)

    @project_db_handler(fall_through=True)
    def handle_projectSaveItems(self, success, request, reply, reason=""):
        data = {'success': success, 'items': reply.get('items', [])}
        broadcast_event(KaraboEvent.ProjectItemsSaved, data)

    @project_db_handler()
    def handle_projectUpdateTrashed(self, success, request, reply,
                                    reason=""):
        if not success:
            reason, details = get_reason_parts(reason)
            messagebox.show_error(reason, details=details)
            return

        data = {"domain": reply["domain"]}
        broadcast_event(KaraboEvent.ProjectTrashed, data)

    def handle_projectUpdate(self, **info):
        """Handle the project update signal from the project manager

        The information is forwarded by the gui server to notify clients about
        updated project items.
        """
        projects = info['info']
        # Get is only required when we connect to an old ecosystem, which will
        # always provide this update msg without client information!
        client_id = projects.get('client', KARABO_CLIENT_ID)
        if client_id == KARABO_CLIENT_ID:
            # If we have been the client saving the project, we are not
            # interested any further!
            return
        uuids = projects['projects']
        broadcast_event(KaraboEvent.ProjectUpdated, {'uuids': uuids})

    # ---------------------------------------------------------------------

    def handle_notification(self, **info):
        """Handle notification events from the GUI server
        """
        message = info.get("message", "")
        content_type = info.get("contentType", "")
        if content_type == "banner":
            broadcast_event(KaraboEvent.ServerNotification, info)
        elif content_type == "logger":
            text = info.get("message", "")
            level = getLevelName(info.get("level", "INFO"))
            get_logger().log(level, text)
        elif message:
            messagebox.show_warning(message)

    def handle_networkData(self, name, data, meta=None):
        """This method handles the big data chunks coming from Karabo
        pipelines to `GuiServerDevice`. To keep the GUI responsive the
        displaying of this data is delayed here.
        """

        def show_data():
            item = self._big_data.pop(name, None)
            if item is None:
                # XXX: This should never happen, but keep going
                get_network().onError(
                    f"Received big data although not scheduled for {name}.")
                get_network().onRequestNetwork(name)
                return
            data_hash, meta_hash = item
            device_id, prop_path = name.split(":")
            device_proxy = self._topology.get_device(device_id)
            if not device_proxy.online:
                return
            binding = device_proxy.get_property_binding(prop_path)
            timestamp = Timestamp.fromHashAttributes(
                meta_hash['timestamp', ...])
            if binding is not None:
                # Note: Binding can be `None` on startup when schema was not
                # arriving and the application closing...
                apply_fast_data(data_hash, binding.value.schema, timestamp)
                device_proxy.config_update = True
                if self._show_big_data_proc:
                    proc = Timestamp().toTimestamp() - timestamp.toTimestamp()
                    info = {'name': name, 'proc': proc}
                    broadcast_event(KaraboEvent.BigDataProcessing, info)
            # Let the GUI server know we have processed this chunk
            get_network().onRequestNetwork(name)

        self._big_data[name] = (data, meta)
        executeLater(show_data, Priority.BIG_DATA)

    def handle_initReply(self, deviceId, success, message):
        device = self._topology.get_device(deviceId)
        if device is not None:
            data = {'device': device, 'success': success, 'message': message}
            # Create KaraboBroadcastEvent
            broadcast_event(KaraboEvent.DeviceInitReply, data)
            if not success:
                reason, details = get_reason_parts(message)
                text = (f'The instance <b>{deviceId}</b> could not be '
                        'instantiated.. <br><br>The reason is:<br>'
                        f'<i>{reason}</i><br>')
                if details is not None:
                    text += '<br>Click "Show Details..." for more information.'
                messagebox.show_error(text, details=details)

    def handle_listDestinations(self, success, request, reply, reason=""):
        """
        Broadcast the available proposals in the Topic, fetched by the
        KaraboLogBook, so that the PreviewDialog can  list them.
        """
        destinations = reply.get("destinations")
        broadcast_event(KaraboEvent.ActiveDestinations, destinations)

    def handle_saveLogBook(self, success, request, reply, reason=""):
        """
        Show messagebox when the KaraboLogBook fails to send the message to
        the logbook.
        """
        if success:
            args = request.get("args")
            data_type = args.get("dataType")
            message = f"Posted the {data_type} to LogBook successfully"
            get_logger().info(message)
        else:
            error, details = get_reason_parts(reason)
            messagebox.show_warning(title="Send to LogBook Failed",
                                    text=error, details=details)

    # ------------------------------------------------------------------
    # Private methods

    def _set_server_information(self, info):
        read_only = info.get('readOnly', False)
        get_network().set_server_information(read_only=read_only)
        broadcast_event(KaraboEvent.ServerInformationUpdate, info)

    # ------------------------------------------------------------------


def _extract_topology_devices(topo_hash):
    """Get all the devices and their classes out of a system topology update.
    """
    devices, servers = [], []

    if 'device' in topo_hash:
        for device_id, _, attrs in topo_hash['device'].iterall():
            # XXX: remove this check after #122877 is fixed on the GUI server
            if not attrs:
                msg = ("skipping malformed topology entry "
                       f"for <b>{device_id}</b>")
                get_logger().error(msg)
                continue
            class_id = attrs['classId']
            devices.append((device_id, class_id, ProxyStatus.ONLINE))

    if 'macro' in topo_hash:
        for device_id, _, attrs in topo_hash['macro'].iterall():
            # XXX: remove this check after #122877 is fixed on the GUI server
            if not attrs:
                msg = ("skipping malformed topology entry "
                       f"for <b>{device_id}</b>")
                get_logger().error(msg)
                continue
            class_id = attrs['classId']
            devices.append((device_id, class_id, ProxyStatus.ONLINE))

    if 'server' in topo_hash:
        for server_id, _, attrs in topo_hash['server'].iterall():
            host = attrs['host']
            servers.append((server_id, host, ProxyStatus.ONLINE))

    return devices, servers



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on March 28, 2018
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from pathlib import Path

from qtpy.QtSvg import QSvgWidget
from traits.api import Instance

from karabo.common.scenemodel.api import GlobalAlarmModel
from karabogui import icons
from karabogui.binding.api import StringBinding, get_binding_value
from karabogui.controllers.api import (
    BaseBindingController, register_binding_controller, with_display_type)
from karabogui.indicators import get_alarm_svg

ICON_PATH = Path(icons.__file__).parent
SVG_OFFLINE = str(ICON_PATH.joinpath("statusOffline.svg"))


@register_binding_controller(ui_name="Alarm Widget",
                             klassname="DisplayAlarm",
                             is_compatible=with_display_type("AlarmCondition"),
                             priority=90,
                             binding_type=StringBinding)
class DisplayAlarm(BaseBindingController):
    # The scene data model class for this controller
    model = Instance(GlobalAlarmModel, args=())

    def create_widget(self, parent):
        widget = QSvgWidget(parent)
        return widget

    def add_proxy(self, proxy):
        # XXX: Only for backward compatibility
        return False

    def clear_widget(self):
        self.widget.load(SVG_OFFLINE)

    def value_update(self, proxy):
        value = get_binding_value(proxy)
        svg = get_alarm_svg(value)
        if svg is None:
            svg = SVG_OFFLINE
        self.widget.load(svg)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QDialog

from karabo.common.scenemodel.api import DisplayListModel
from karabo.native import Configurable, VectorInt32
from karabogui.binding.api import apply_default_configuration
from karabogui.controllers.display.list import DisplayList
from karabogui.fonts import get_font_size_from_dpi
from karabogui.testing import get_class_property_proxy, set_proxy_value


class Object(Configurable):
    prop = VectorInt32(defaultValue=[1])


def test_list_controller(gui_app, mocker):
    proxy = get_class_property_proxy(Object.getClassSchema(), "prop")
    controller = DisplayList(proxy=proxy,
                             model=DisplayListModel())
    controller.create(None)
    apply_default_configuration(proxy.root_proxy.binding)
    assert controller.widget.focusPolicy() == Qt.NoFocus

    set_proxy_value(proxy, "prop", [0, 2])
    assert controller.widget.text() == "0,2"

    model = controller.model
    assert model.font_size == 10
    assert model.font_weight == "normal"

    action = controller.widget.actions()[0]
    assert action.text() == "Format field..."

    path = "karabogui.controllers.display.list.FormatLabelDialog"
    dialog = mocker.patch(path)
    dialog().exec.return_value = QDialog.Accepted
    dialog().font_size = 12
    dialog().font_weight = "bold"
    action.trigger()
    assert dialog().exec.call_count == 1
    assert controller.model.font_size == 12
    assert controller.model.font_weight == "bold"

    widget = controller.widget
    controller._apply_format(widget)
    font = widget.font()
    size = font.pointSize()
    bold = font.bold()

    assert size == get_font_size_from_dpi(controller.model.font_size)
    assert bold == (controller.model.font_weight == "bold")

    controller.destroy()
    assert controller.widget is None



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import datetime

import numpy
from pyqtgraph import PlotDataItem
from qtpy.QtCore import QDateTime
from traits.api import (
    Array, Constant, Float, HasStrictTraits, Instance, Int, List, WeakRef,
    on_trait_change)

from karabo.native import Timestamp
from karabogui.binding.api import PropertyProxy
from karabogui.graph.common.const import ALARM_INTEGER_MAP, STATE_INTEGER_MAP

ONE_WEEK = "One Week"
ONE_DAY = "One Day"
ONE_HOUR = "One Hour"
TEN_MINUTES = "Ten Minutes"
UPTIME = "Uptime"
HIDDEN = "Hidden"


def get_start_end_date_time(selected_time_span):
    """ Return beginning and end date time for given ``selected_time_span``.
        If the ``selected_time_span`` is not supported ``None`` is returned.
    """
    current_date_time = QDateTime.currentDateTime()
    if selected_time_span == ONE_WEEK:
        # One week
        start_date_time = current_date_time.addDays(-7)
    elif selected_time_span == ONE_DAY:
        # One day
        start_date_time = current_date_time.addDays(-1)
    elif selected_time_span == ONE_HOUR:
        # One hour
        start_date_time = current_date_time.addSecs(-3600)
    elif selected_time_span == TEN_MINUTES:
        # Ten minutes
        start_date_time = current_date_time.addSecs(-600)
    else:
        return None, None

    return start_date_time, current_date_time


class _Generation:
    """ This holds a single generation of a Curve's data.
    """
    size = 200
    base = 10  # number of points to combine per generation

    def __init__(self):
        self.fill = 0
        self.xs = numpy.empty(self.size, dtype=float)
        self.ys = numpy.empty(self.size, dtype=float)

    def add_point(self, x, y):
        self.xs[self.fill] = x
        self.ys[self.fill] = y
        self.fill += 1

        if self.fill == self.size:
            return self.reduce_data()
        return None

    def reduce_data(self):
        x = self.xs[:self.base].mean()
        y = self.ys[:self.base].mean()
        self.xs[:-self.base] = self.xs[self.base:]
        self.ys[:-self.base] = self.ys[self.base:]
        self.fill -= self.base
        return x, y


class Curve(HasStrictTraits):
    """This holds the data for one curve

    The currently to be shown data is in self.x and self.y.
    Up to self.histsize it is filled with historical data, up to self.fill
    it is then filled with "current" data, meaning data that has been
    accumulated from the changes coming in.

    There is a second data structure, self.generations, which is a list of
    _Generation objects containing averaged data, always `base` points are
    averaged over and put into the next higher aggregated storage.

    Once the basic storage in self.x and self.y flows over, it gets replaced
    by the averaged data in self.generations.
    """
    curve = WeakRef(PlotDataItem)
    proxy = Instance(PropertyProxy)
    generations = List(Instance(_Generation))

    histsize = Int(0)
    fill = Int(0)

    x = Array(dtype=float)
    y = Array(dtype=float)

    t0 = Float(0)
    t1 = Float(0)

    spare = Constant(100)
    # Limits amount of data from past
    maxHistory = Constant(500)
    sparsesize = Constant(400)
    # minimum number of points shown (if possible)
    minHistory = Constant(100)

    # Defines which curve we are using. Curve type 0 is classic, 1 is state
    # and 2 is alarms
    curve_type = Int(0)

    def _generations_default(self):
        return [_Generation() for _ in range(4)]

    def _x_default(self):
        arraysize = self.spare + sum([g.size for g in self.generations], 0)
        return numpy.empty(arraysize, dtype=float)

    def _y_default(self):
        arraysize = self.spare + sum([g.size for g in self.generations], 0)
        return numpy.empty(arraysize, dtype=float)

    def purge(self):
        """Purge the curve by resetting all values to their default"""
        self.generations = self._generations_default()
        self.x = self._x_default()
        self.y = self._y_default()
        self.histsize = 0
        self.fill = 0
        self.update()

    def add_point(self, value, timestamp):
        # Fill the generations data, possibly propagating averaged values
        point = (timestamp, value)
        for gen in reversed(self.generations):
            point = gen.add_point(*point)
            if point is None:
                break

        # Fill the main data buffer
        self.x[self.fill] = timestamp
        self.y[self.fill] = value
        self.fill += 1
        # When the main buffer fills up, copy the generations data
        if self.fill == len(self.x):
            self.fill_current()
        self.update()

    def fill_current(self):
        pos = self.histsize
        for gen in self.generations:
            fill = gen.fill
            if fill == 0:
                continue
            self.x[pos:pos + fill] = gen.xs[:fill]
            self.y[pos:pos + fill] = gen.ys[:fill]
            pos += fill
        self.fill = pos

    def get_property_history(self, t0, t1):
        # Avoid if not currently visible
        if not self.proxy.visible:
            return
        t0 = str(datetime.datetime.utcfromtimestamp(t0).isoformat())
        t1 = str(datetime.datetime.utcfromtimestamp(t1).isoformat())
        self.proxy.get_history(t0, t1, max_value_count=self.maxHistory)

    def changeInterval(self, t0, t1, force=False):
        """Change the time interval of this Curve

        :param t0: new start time of the curve
        :param t1: new end time of the curve
        """
        p0 = self.x[:self.fill].searchsorted(t0)
        p1 = self.x[:self.fill].searchsorted(t1)

        not_enough_data = (p1 - p0) < self.minHistory
        no_data = self.histsize == 0
        zoomed_out = self.histsize > self.sparsesize
        nearly_left_border = 0.9 * self.t0 + 0.1 * self.t1
        nearly_right_border = 0.1 * self.t0 + 0.9 * self.t1

        # Request history only needs to be done under certain circumstances
        if (force or no_data or (not_enough_data and zoomed_out) or
                (self.x[p0] > nearly_left_border) or
                (p1 < self.histsize and self.x[p1 - 1] < nearly_right_border)):
            self.get_property_history(t0, t1)

        self.t0 = t0
        self.t1 = t1

    def update(self):
        """ Show the new data on screen"""
        if self.curve:
            self.curve.setData(self.x[:self.fill], self.y[:self.fill])

    @on_trait_change("proxy:visible")
    def _visibility_update(self, visible):
        if visible and self.t1 >= self.x[self.histsize]:
            self.get_property_history(self.t0, self.t1)

    @on_trait_change("proxy:binding:historic_data")
    def _historic_data_arrival(self, data):
        if not data:
            return

        if self.t1 == self.t0:
            # Prevent division by 0, otherwise self.curve.plot() is None
            return

        datasize = len(data)
        gensize = sum([g.size for g in self.generations], 0)
        arraysize = datasize + gensize + self.spare
        x = numpy.empty(arraysize, dtype=float)
        y = numpy.empty(arraysize, dtype=float)

        for i, d in enumerate(data):
            # Protect against inf by setting numpy.NaN which is not shown
            x[i] = Timestamp.fromHashAttributes(d["v", ...]).toTimestamp()
            # Do some gymnastics for crazy values!
            value = d["v"]
            if self.curve_type == 1:
                value = STATE_INTEGER_MAP[value]
            elif self.curve_type == 2:
                value = ALARM_INTEGER_MAP[value]
            y[i] = value

        p0 = self.x[:self.fill].searchsorted(self.t0)
        p1 = self.x[:self.fill].searchsorted(self.t1)
        np0 = x[:datasize].searchsorted(self.t0)
        np1 = x[:datasize].searchsorted(self.t1)

        span = (self.x[p1 - 1] - self.x[p0]) / (self.t1 - self.t0)
        nspan = (x[np1 - 1] - x[np0]) / (self.t1 - self.t0)

        if (np1 - np0 < p1 - p0) and not nspan > span < 0.9:
            return

        # If the history overlaps generation data, favor the history data.
        end = x[datasize - 1]
        for gen in self.generations:
            fill = gen.fill
            if fill == 0:
                continue
            pos = gen.xs[:fill].searchsorted(end)
            if pos == 0:
                break
            gen.xs[:fill - pos] = gen.xs[pos:fill]
            gen.ys[:fill - pos] = gen.ys[pos:fill]
            gen.fill -= pos

        self.histsize = datasize
        self.x = x
        self.y = y
        self.fill_current()
        self.update()

    def get_max_timepoint(self):
        """ Return the max time point for the fill values"""
        return numpy.max(self.x[:self.fill]) if self.fill else 0.0



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from collections import namedtuple

from qtpy import uic
from qtpy.QtCore import (
    QAbstractTableModel, QItemSelection, QModelIndex, QPoint,
    QSortFilterProxyModel, Qt, Slot)
from qtpy.QtGui import QPalette
from qtpy.QtWidgets import (
    QDialog, QDialogButtonBox, QHeaderView, QMenu, QMessageBox)

import karabogui.icons as icons
from karabogui import messagebox
from karabogui.access import (
    AccessRole, access_role_allowed, get_access_level_for_role)
from karabogui.events import (
    KaraboEvent, broadcast_event, register_for_broadcasts,
    unregister_from_broadcasts)
from karabogui.singletons.api import get_network, get_topology
from karabogui.util import utc_to_local
from karabogui.validators import RegexValidator

from .utils import get_dialog_ui

NAME_COLUMN = 0


class SaveConfigurationDialog(QDialog):
    """Save dialog for configuration details"""

    def __init__(self, parent=None):
        super().__init__(parent)
        uic.loadUi(get_dialog_ui("init_config_save.ui"), self)
        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(False)
        self.ui_name.textChanged.connect(self._check_button)
        # We allow only characters and numbers up to a length of 30!
        validator = RegexValidator(pattern="^[A-Za-z0-9_-]{1,30}$")
        self.ui_name.setValidator(validator)

        self._normal_palette = self.ui_name.palette()
        self._error_palette = QPalette(self._normal_palette)
        self._error_palette.setColor(QPalette.Text, Qt.red)

        self.event_map = {
            KaraboEvent.NetworkConnectStatus: self._event_network
        }
        register_for_broadcasts(self.event_map)

    @Slot()
    def _check_button(self):
        """Basic validation of configuration name"""
        enable = self.ui_name.hasAcceptableInput()
        if enable:
            self.ui_name.setPalette(self._normal_palette)
        else:
            self.ui_name.setPalette(self._error_palette)
        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(enable)

    def _event_network(self, data):
        if not data.get("status"):
            self.close()

    def done(self, result):
        """Stop listening for broadcast events"""
        unregister_from_broadcasts(self.event_map)
        super().done(result)

    @property
    def name(self):
        return self.ui_name.text()


class InitConfigurationDialog(QDialog):
    """List configurations by ``name`` from the configuration database
    """

    def __init__(self, instance_id, parent=None):
        super().__init__(parent)
        uic.loadUi(get_dialog_ui("init_configuration_dialog.ui"), self)
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.setModal(False)
        self.setWindowFlags(self.windowFlags() | Qt.WindowCloseButtonHint)

        self.instance_id = instance_id
        self.ui_instance.setText(f"Device Id: {instance_id}")

        self._server_id = None
        self._class_id = None
        self.event_map = {
            KaraboEvent.ListConfigurationUpdated: self._event_list_updated,
            KaraboEvent.NetworkConnectStatus: self._event_network,
            KaraboEvent.AccessLevelChanged: self._event_access_level,
        }
        register_for_broadcasts(self.event_map)

        self.model = QSortFilterProxyModel(parent=self)
        self.model.setSourceModel(TableModel(parent=self))
        self.model.setFilterRole(Qt.DisplayRole)
        self.model.setFilterFixedString("")
        self.model.setFilterCaseSensitivity(False)
        self.model.setFilterKeyColumn(0)

        self.ui_table_widget.setModel(self.model)
        header = self.ui_table_widget.horizontalHeader()
        header.setSectionResizeMode(QHeaderView.ResizeToContents)
        header.setStretchLastSection(True)
        self.ui_table_widget.setContextMenuPolicy(Qt.CustomContextMenu)
        self.ui_table_widget.selectionModel().selectionChanged.connect(
            self._selectionChanged)
        self.ui_table_widget.doubleClicked.connect(self._request_configuration)
        self.ui_table_widget.customContextMenuRequested.connect(
            self.onCustomContextMenuRequested)

        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(False)
        self.ui_filter.textChanged.connect(self._filter_changed)
        self.ui_button_clear.clicked.connect(self.ui_filter.clear)

        self.ui_filter.setFocus()

        self.ui_button_save.setIcon(icons.save)
        self.ui_show_device.setIcon(icons.deviceInstance)
        self.ui_button_refresh.setIcon(icons.reset)
        # Provide saving option
        self.ui_button_save.clicked.connect(self.open_save_dialog)

        self.ui_button_request.clicked.connect(self._request_configuration)

        # Show device option!
        self.ui_show_device.clicked.connect(self._show_device)

        self.ui_button_delete.setIcon(icons.delete)
        self.ui_button_delete.clicked.connect(self._request_delete)

        # Request fresh at startup!
        get_network().onListInitConfigurations(device_id=self.instance_id)

    def _event_list_updated(self, data):
        instance_id = data["deviceId"]
        if instance_id != self.instance_id:
            return

        model = self.model.sourceModel()
        items = data["items"]
        if not items:
            self.ui_status.setText("No configurations are available!")
        else:
            self.ui_status.setText("")
            self._server_id = data["items"][0]["serverId"]
            self._class_id = data["items"][0]["classId"]
        model.set_internal_data(items)
        self._check_button_state()

    def _event_network(self, data):
        if not data.get("status"):
            self.close()

    def _event_access_level(self, data):
        """Update the enabled status of Delete button on access level
        changes"""
        allow_deletion = access_role_allowed(AccessRole.CONFIGURATION_DELETE)
        enable = allow_deletion and self._check_existing()
        self.ui_button_delete.setEnabled(enable)

    def done(self, result):
        """Stop listening for broadcast events"""
        unregister_from_broadcasts(self.event_map)
        super().done(result)

    def _check_existing(self):
        if self.ui_table_widget.selectionModel().hasSelection():
            return True

        return False

    def _check_button_state(self):
        enable = self._check_existing()
        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(enable)
        self.ui_button_request.setEnabled(enable)
        enable = enable and access_role_allowed(
            AccessRole.CONFIGURATION_DELETE)
        self.ui_button_delete.setEnabled(enable)

    # --------------------------------------------------------------------
    # Qt Slots

    @Slot(QPoint)
    def onCustomContextMenuRequested(self, pos):
        rows = self.ui_table_widget.selectionModel().selectedRows()
        if not rows:
            return

        menu = QMenu(parent=self.ui_table_widget)
        delete_action = menu.addAction(icons.delete, "Delete")
        delete_action.triggered.connect(self._request_delete)
        allowed = access_role_allowed(AccessRole.CONFIGURATION_DELETE)
        if not allowed:
            delete_action.setEnabled(False)
            level = get_access_level_for_role(AccessRole.CONFIGURATION_DELETE)
            delete_action.setToolTip(
                f"Need access level {level} to delete configuration.")
        menu.exec(self.ui_table_widget.viewport().mapToGlobal(pos))

    @Slot()
    def _request_delete(self):
        index = self.ui_table_widget.selectionModel().selectedRows()[0]
        if not index.isValid():
            return
        model = index.model()
        name = model.index(index.row(), NAME_COLUMN).data()

        confirm = QMessageBox(
            QMessageBox.Question, "Delete Configuration?",
            f"Do you really want to delete the configuration '{name}'?",
            QMessageBox.No | QMessageBox.Yes, parent=self).exec()

        if confirm == QMessageBox.No:
            return

        get_network().onDeleteInitConfiguration(
            self.instance_id, name)

    @Slot()
    def _show_device(self):
        broadcast_event(KaraboEvent.ShowDevice, {'deviceId': self.instance_id,
                                                 'showTopology': True})

    @Slot()
    def open_save_dialog(self):
        dialog = SaveConfigurationDialog(parent=self)
        if dialog.exec() == QDialog.Accepted:
            device = get_topology().get_device(self.instance_id)
            if not device.online:
                text = (f"The device {self.instance_id} is not online, the "
                        f"configuration cannot be saved.")
                messagebox.show_alarm(text, parent=self)
                return
            name = dialog.name

            # Check if the name already exists.
            model = self.model.sourceModel()
            first_index = model.index(0, 0, QModelIndex())
            hits = 1
            if model.match(first_index, Qt.DisplayRole, name, hits,
                           Qt.MatchCaseSensitive):
                ask = (f"A configuration with the name '{name}' already "
                       "exists. Do you wish to overwrite it ? ")
                overwrite = QMessageBox(
                    QMessageBox.Question, "Overwrite Configuration?",
                    ask, QMessageBox.No | QMessageBox.Yes,
                    parent=self).exec()
                if overwrite == QMessageBox.No:
                    return

            get_network().onSaveInitConfiguration(
                name, [self.instance_id], update=True)

    @Slot()
    def accept(self):
        """The dialog was accepted and we can request a configuration"""
        self._request_configuration()
        super().accept()

    @Slot()
    def _request_configuration(self):
        """Request the configuration for the actual `instance_id`"""
        index = self.ui_table_widget.selectionModel().selectedRows()[0]
        if not index.isValid():
            return
        model = index.model()
        name = model.index(index.row(), NAME_COLUMN).data()
        preview = self.ui_preview.isChecked()
        get_network().onGetInitConfiguration(
            self.instance_id, name, preview)

    @Slot()
    def on_ui_button_refresh_clicked(self):
        """Refresh the list of configurations for the actual `instance_id`"""
        get_network().onListInitConfigurations(self.instance_id)

    @Slot(str)
    def _filter_changed(self, text):
        self.model.setFilterFixedString(text)
        self._check_button_state()

    @Slot(QItemSelection, QItemSelection)
    def _selectionChanged(self, selected, deselected):
        self._check_button_state()


ConfigurationEntry = namedtuple("ConfigurationEntry", ["name", "date"])


class TableModel(QAbstractTableModel):
    headers = ["Name", "Date"]

    def __init__(self, instance_id="", parent=None):
        super().__init__(parent)
        self.instance_id = instance_id
        self.data = []

    def set_internal_data(self, data):
        """ Add the given `data` to the internal data structure
        """
        self.beginResetModel()
        try:
            self.data = [
                ConfigurationEntry(
                    name=str(item["name"]),
                    date=utc_to_local(item["date"]),
                ) for item in data]
        finally:
            self.endResetModel()

    def rowCount(self, parent=None):
        return len(self.data)

    def columnCount(self, parent=None):
        return len(self.headers)

    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid():
            return None

        if role in (Qt.DisplayRole, Qt.ToolTipRole):
            entry = self.data[index.row()]
            column = index.column()
            return entry[column]

    def headerData(self, section, orientation, role):
        if role == Qt.DisplayRole and orientation == Qt.Horizontal:
            return self.headers[section]
